"""
Project: kiltera 
Module: event_sched
Author: Ernesto Posse
Created on: Sep 25, 2006
Last modified: Jan 22, 2007

Description:
  This implements an event-based simulator of kiltera processes.
  It is based on the scheduler class of the Python standard library.
"""

import bisect
import random
from kexceptions import *
from kiltera_ast import *
from posse_tools.utils import indent
import time
import pprint

pp = pprint.PrettyPrinter(indent = 4)

# The constant EPSILON is the margin of error for time, i.e. the maximum difference allowed between two times for them to be considered equal.
EPSILON = 0.00001

class EventScheduler:
  
    def __init__(self, clock):
        self.queue = []
        self.clock = clock
        self.timefunc = clock.get_time
        self.delayfunc = clock.delay

    def enterabs(self, time, priority, target, *args, **kwargs):
        """Enter a new event in the queue at an absolute time.

        Returns an ID for the event which can be used to remove it,
        if necessary.
        """
        event = Event(time, priority, target, *args, **kwargs)
        new_time_slot = TimeSlot(time, [event])
        i = bisect.bisect(self.queue, new_time_slot)
        if i > 0 and abs(self.queue[i-1].time - time) <= EPSILON:
            self.queue[i-1].insert(event)
        else:
            bisect.insort(self.queue, new_time_slot)
        log.note(4, "scheduled new event:"+repr(event))
        return event # The ID

    def enterabs_first(self, time, target, *args, **kwargs):
        """Enter a new event in the queue at an absolute time with the 
        highest priority for that time-slot.

        Returns an ID for the event which can be used to remove it,
        if necessary.
        """
        priority = 0
        event = Event(time, priority, target, *args, **kwargs)
        new_time_slot = TimeSlot(time, [event])
        i = bisect.bisect(self.queue, new_time_slot)
        if i > 0 and abs(self.queue[i-1].time - time) <= EPSILON:
            if not self.queue[i-1].isempty():
                current_first_priority = self.queue[i-1].first().priority
                event.priority = current_first_priority - 1
            self.queue[i-1].insertat(0, event)
        else:
            bisect.insort(self.queue, new_time_slot)
        log.note(4, "scheduled new event:"+repr(event))
        return event # The ID

    def enterabs_last(self, time, target, *args, **kwargs):
        """Enter a new event in the queue at an absolute time with the 
        lowest priority for that time-slot.

        Returns an ID for the event which can be used to remove it,
        if necessary.
        """
        priority = 0
        event = Event(time, priority, target, *args, **kwargs)
        new_time_slot = TimeSlot(time, [event])
        i = bisect.bisect(self.queue, new_time_slot)
        if i > 0 and abs(self.queue[i-1].time - time) <= EPSILON:
            if not self.queue[i-1].isempty():
                current_last_priority = self.queue[i-1].last().priority
                event.priority = current_last_priority + 1
            self.queue[i-1].append(event)
        else:
            bisect.insort(self.queue, new_time_slot)
        log.note(4, "scheduled new event:"+repr(event))
        return event # The ID

    def enterabs_rand(self, time, target, *args, **kwargs):
        """Enter a new event in the queue at an absolute time with a 
        random priority for that time-slot.

        Returns an ID for the event which can be used to remove it,
        if necessary.
        """
        priority = 0
        event = Event(time, priority, target, *args, **kwargs)
        new_time_slot = TimeSlot(time, [event])
        i = bisect.bisect(self.queue, new_time_slot)
        if i > 0 and abs(self.queue[i-1].time - time) <= EPSILON:
            if not self.queue[i-1].isempty():
                current_last_priority = self.queue[i-1].last().priority
                current_first_priority = self.queue[i-1].first().priority
                event.priority = random.randint(current_first_priority, current_last_priority)
            self.queue[i-1].insert(event)
        else:
            bisect.insort(self.queue, new_time_slot)
        log.note(4, "scheduled new event:"+repr(event))
        return event # The ID

    def enter(self, delay, priority, target, *args, **kwargs):
        """A variant that specifies the time as a relative time.

        This is actually the more commonly used interface.
        """
        time = self.timefunc() + delay
        return self.enterabs(time, priority, target, *args, **kwargs)

    def enter_first(self, delay, target, *args, **kwargs):
        """A variant that specifies the time as a relative time.

        This is actually the more commonly used interface.
        """
        time = self.timefunc() + delay
        return self.enterabs_first(time, target, *args, **kwargs)

    def enter_last(self, delay, target, *args, **kwargs):
        """A variant that specifies the time as a relative time.

        This is actually the more commonly used interface.
        """
        time = self.timefunc() + delay
        return self.enterabs_last(time, target, *args, **kwargs)

    def enter_rand(self, delay, target, *args, **kwargs):
        """A variant that specifies the time as a relative time.

        This is actually the more commonly used interface.
        """
        time = self.timefunc() + delay
        return self.enterabs_rand(time, target, *args, **kwargs)

    def cancel(self, event):
        """Remove an event from the queue.

        This must be presented the ID as returned by enter().
        If the event is not in the queue, this raises RuntimeError.
        """
        i = bisect.bisect(self.queue, TimeSlot(event.time, [event]))
        if i <= 0 or abs(self.queue[i-1].time - event.time) > EPSILON:
            raise SimulatorException("Attempting to cancel an event that is not scheduled: no time-slot exists for that event: %s" % event)
        self.queue[i-1].remove(event)
        if self.queue[i-1].isempty():
            del self.queue[i-1]
        log.note(4, "canceled event:"+repr(event))

    def isempty(self):
        """Check whether the queue is empty."""
        return len(self.queue) == 0

    def run(self):
        """Execute events until the queue is empty.
        """
        q = self.queue
        while q:
            now = self.timefunc()
            log.note(3, 'current time = %f' % now)
            time_slot = q[0]
            time = time_slot.time
            event_list = time_slot.event_list
            log.note(3, 'next time-slot = %f' % time)
            if now < time:
                log.note(3, 'advancing to next time slot by %f time units' % (time - now))
                self.delayfunc(time - now)
            else:
                log.note(3, 'executing time-slot\n  event list = \n%s' % pp.pformat(event_list))
                while event_list:
                    event = event_list[0]
                    log.note(3, 'executing new event\n  %s' % pp.pformat(event))
                    del event_list[0]
                    event()
                    time.sleep(0)   # Let other threads run
                del q[0]
    

class TimeSlot:
    
    def __init__(self, time, event_list=None):
        self.time = time
        if event_list is None:
            self.event_list = []
        else:
            self.event_list = event_list
    
    def __cmp__(self, other):
        return cmp(self.time, other.time)
    
    def isempty(self):
        return len(self.event_list) == 0
        
    def first(self):
        return self.event_list[0]
    
    def last(self):
        return self.event_list[-1]
    
    def append(self, event):
        self.event_list.append(event)
    
    def insert(self, event):
        bisect.insort(self.event_list, event)
    
    def insertat(self, index, event):
        self.event_list.insert(index, event)
    
    def remove(self, event):
        self.event_list.remove(event)

    def __str__(self):
        s = 'time = '+str(self.time)+'\n'
        for e in self.event_list:
            s += '  '+str(e)+'\n'
        return s
    
    def __repr__(self):
        return "TimeSlot(" + repr(self.time) \
            +  ", \n" + indent(pp.pformat(self.event_list),4) \
            +  ")"


class Event:
  
    def __init__(self, time, priority, target, *args, **kwargs):
        self.time = time
        self.priority = priority
        self.target = target
        self.args = args
        self.kwargs = kwargs
        if 'addtoprocess' in kwargs and kwargs['addtoprocess'] \
        or not ('addtoprocess' in kwargs):
            if isinstance(self.target, ProcessInstance):
                self.target.add_event(self)
    
    def __call__(self):
        if isinstance(self.target, ProcessInstance):
            self.target.remove_event(self)
        self.target(*self.args, **self.kwargs)
    
    def __cmp__(self, other):
        #~ if self.time == other.time:
        if abs(self.time - other.time) <= EPSILON:
            return cmp(self.priority, other.priority)
        return cmp(self.time, other.time)
    
    def __str__(self):
        return "event time = " + str(self.time) \
            +  " pri = " + str(self.priority) \
            +  " target = " + repr(self.target) \
            +  " args = " + repr(self.args) \
            +  " kwargs = " + repr(self.kwargs)
    
    def __repr__(self):
        return "Event(" + repr(self.time) \
            +  ", " + str(self.priority) \
            +  ", " + repr(self.target) \
            +  ", *" + repr(self.args) \
            +  ", **" + repr(self.kwargs) \
            +  ")"

