"""
Project: kiltera 
Module: interpreter
Author: Ernesto Posse
Created on: Jun 9, 2006
Last modified: Jan 22, 2007

Description:
  This implements an interpreter for the execution of kiltera processes.
"""

from kiltera_ast import *
from values import *
from environments import *
from channels import *
from kexceptions import *
from threading import *
from ktime import *
from kthreads import *
from visitor import *
from posse_tools.utils import *
from builtins import *
import socket

HOST = socket.gethostname()
PORT = 50007

class Interpreter(Visitor):
    
    def __init__(self, server=None, port=PORT):
        self.observers = []
        self.clock = PhysicalClock()
        self.server = server
        self.port = port
    
    def execute(self, module, main=None):
        log.note(1, "1 -- execution begins")
        new_env = Env(Frame(self.builtins()))
        if self.server is not None:
            self.setup_server_connection(module)
        self.main_thread = KThread(target = module, args=(self, new_env), kwargs={'main':main})
        self.keyboard_handler = KeyboardInterruptHandler(target=self.main_thread)
        self.main_thread.keyboard_handler = self.keyboard_handler
        try:
            self.main_thread.start()
            self.keyboard_handler.run()
            if self.server is not None:
                self.close_server_connection()
        except RuntimeException, e:
            print e
        log.note(1, "2 -- execution ends")

    def setup_server_connection(self, module):
        name = module.name
        log.note(1, "setup_server_connection 1")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        log.note(1, "setup_server_connection 2")
        self.socket.connect((self.server, self.port))
        log.note(1, "setup_server_connection 3")
        self.socket.send(name)
        log.note(1, "setup_server_connection 4")
        ack = self.socket.recv(1024)
        log.note(1, "setup_server_connection 5")
        if ack == 'invalid name':
            raise RuntimeException(msg='invalid module name')
        elif ack != 'acknowledged':
            raise RuntimeException(msg='bad server acknowledgement')
        log.note(1, "setup_server_connection 6")
        
    def close_server_connection(self):
        log.note(1, "close_server_connection 1")
        self.socket.close()
        log.note(1, "close_server_connection 2")

    def register(self, observer):
        self.observers.append(observer)
    
    def update(self, *args, **kwargs):
        for observer in self.observers:
            observer.notify(*args, **kwargs)
    
    # Atom visitors
    
    def visit_atom(self, atom, *args, **kwargs):
        env = args[0]
        
    def visit_output(self, output, *args, **kwargs):
        env = args[0]
        try:
            channel = output.channel_ref.accept(self, env)
        except EnvironmentException:
            raise RuntimeException(output.parent, "incorrect channel reference in send")
        currentThread().register_blocking_resource(channel)
        data = output.data.accept(self, env)
        channel.send(data)
        self.update(time=self.clock.time, kind='ext', action='send', channel=output.channel_ref, data=data, loc=output.parent.get_full_location(), pos=output.parent.pos)
        
    def visit_input(self, input, *args, **kwargs):
        env = args[0]
        channel = input.channel_ref.accept(self, env)
        currentThread().register_blocking_resource(channel)
        try:
            data = channel.receive()
        except EOFError:
            raise RuntimeException(input, "Remote connection lost")
        self.update(time=self.clock.time, kind='ext', action='receive', channel=input.channel_ref, data=data, loc=input.parent.get_full_location(), pos=input.parent.pos)
        return data
        
    def visit_assign(self, assignment, *args, **kwargs):
        env = args[0]
        try:
            actual_value = assignment.expression.accept(self, env)
            env[assignment.variable] = actual_value
            self.update(time=self.clock.time, kind='int/ext', action='assign', var=assignment.variable, data=actual_value, loc=assignment.parent.get_full_location(), pos=assignment.parent.pos)
        except EnvironmentException:
            raise RuntimeException(assignment.parent, "assignment error") 
        
    def visit_print(self, print_atom, *args, **kwargs):
        env = args[0]
        msg = print_atom.msg.accept(self, env)
        print str(msg)
        self.update(time=self.clock.time, kind='int', action='print', data=msg, loc=print_atom.parent.get_full_location(), pos=print_atom.parent.pos)
        
    
    
    # Process visitors
    
    def visit_process(self, process, *args, **kwargs):
        env = args[0]
        
    def visit_nil(self, nil, *args, **kwargs):
        env = args[0]
        log.note(2, "1 -- %s", nil)
        deadlock = KEvent()
        deadlock.wait()

    def visit_done(self, done, *args, **kwargs):
        env = args[0]
        log.note(2, "1 -- %s", done)

    def visit_prefix(self, prefix, *args, **kwargs):
        log.note(2, "1 -- %s", prefix)
        currentThread().check_life()
        log.note(2, "2 -- %s", prefix)
        env = args[0]
        prefix.action.accept(self, env)
        log.note(2, "3 -- %s", prefix)
        prefix.continuation.accept(self, *args, **kwargs)
        
    def visit_send(self, send, *args, **kwargs):
        log.note(2, "1 -- %s", send)
        me = currentThread()
        me.check_life()
        log.note(2, "2 -- %s", send)
        env = args[0]
        #~ send.action.path = send.get_full_location()
        send.action.accept(self, env)
        log.note(2, "3 -- %s", send)
        if hasattr(me, 'signal'):
            log.note(2, "4 -- %s", send)
            me.signal.set()
        send.continuation.accept(self, *args, **kwargs)
        
    def visit_receive(self, receive, *args, **kwargs):
        log.note(2, "1 -- %s", receive)
        me = currentThread()
        me.check_life()
        log.note(2, "2 -- %s", receive)
        env = args[0]
        data = receive.action.accept(self, env)
        log.note(2, "3 -- %s", receive)
        if hasattr(me, 'signal'):
            log.note(2, "4 -- %s", receive)
            me.signal.set()
        new_env = Env(Frame({receive.parameter: data}), env)
        receive.continuation.accept(self, new_env, *args[1:], **kwargs)
        
    def visit_parallel(self, parallel, *args, **kwargs):
        log.note(2, "1 -- %s", parallel)
        me = currentThread()
        me.check_life()
        log.note(2, "2 -- %s", parallel)
        env = args[0]
        threads = [KThread(target = process, args = (self, env) + args, kwargs = kwargs) 
                   for process in parallel.process_list]
        me.subthreads += threads
        if hasattr(me, 'signal'):
            log.note(2, "3 -- %s", parallel)
            for thread in threads:
                thread.signal = me.signal
        log.note(2, "4 -- %s\n\n%s\n", parallel,self.main_thread.tree_str())
        for thread in threads:
            thread.start()
        self.join(threads)
        log.note(2, "5 -- %s", parallel)
        
    def visit_sequence(self, sequence, *args, **kwargs):
        log.note(2, "1 -- %s", sequence)
        currentThread().check_life()
        log.note(2, "2 -- %s", sequence)
        env = args[0]
        for process in sequence.process_list:
            log.note(2, "3 -- %s", sequence)
            #~ currentThread().check_life()
            value = process.accept(self, *args, **kwargs)
            #~ if isinstance(value, ExecutingProcess):
                #~ self.join(value.threads)
        log.note(2, "4 -- %s", sequence)

    def visit_local(self, local, *args, **kwargs):
        log.note(2, "1 -- %s", local)
        currentThread().check_life()
        log.note(2, "2 -- %s", local)
        env = args[0]
        if local.channel_class is Sync:
            channel_class = SyncChannel
        else:
            channel_class = AsyncChannel
        table = dict((name, channel_class()) for name in local.channel_names)
        #~ table = dict((name, channel_class(name, local.get_full_location())) for name in local.channel_names)
        new_env = Env(Frame(table), env)
        return local.process.accept(self, new_env, *args[1:], **kwargs)

    def visit_chanarray(self, chanarray, *args, **kwargs):
        log.note(2, "1 -- %s", chanarray)
        currentThread().check_life()
        env = args[0]
        name = chanarray.name
        try:
            size = chanarray.size.accept(self, env)
            if chanarray.channel_class is Sync:
                channel_class = SyncChannel
            else:
                channel_class = AsyncChannel
            array = [channel_class() for i in range(size)]
        except TypeError, e:
            raise RuntimeException(chanarray, "channel array size is of incorrect type: %s" % size.__class__.__name__)
        new_binding = dict([(name, Tuple(array))])
        new_env = Env(Frame(new_binding), env)
        return chanarray.process.accept(self, new_env, *args[1:], **kwargs)

    def visit_let(self, let, *args, **kwargs):
        log.note(2, "1 -- %s", let)
        currentThread().check_life()
        env = args[0]
        bindings = dict((var, let.bindings[var].accept(self, env)) 
                        for var in let.bindings)
        new_env = Env(Frame(bindings), env)
        return let.process.accept(self, new_env, *args[1:], **kwargs)

    def visit_conditional(self, cond, *args, **kwargs):
        log.note(2, "1 -- %s", cond)
        currentThread().check_life()
        env = args[0]
        condition_value = cond.condition.accept(self, env).value
        if condition_value is True:
            return cond.action.accept(self, *args, **kwargs)
        else:
            return cond.alternative.accept(self, *args, **kwargs)

    def visit_match(self, match_proc, *args, **kwargs):
        log.note(2, "1 -- %s", match_proc)
        currentThread().check_life()
        env = args[0]
        expr_val = match_proc.expression.accept(self, env)
        for pattern, process in match_proc.cases:
            corresponds, subst = match(pattern, expr_val)
            if corresponds:
                new_env = Env(Frame(subst), env)
                return process.accept(self, new_env, *args[1:], **kwargs)
        return Bottom()
        
    def visit_procinst(self, process_inst, *args, **kwargs):
        log.note(2, "1 -- %s", process_inst)
        currentThread().check_life()
        env = args[0]
        try:
            latent_process = env[process_inst.process_name]
            processdef = latent_process.processdef
        except EnvironmentException:
            raise RuntimeException(process_inst, "process definition '%s' not found in the current environment" % process_inst.process_name)
        except AttributeError:
            raise RuntimeException(process_inst, "'%s' is not a valid process" % process_inst.process_name)
        actual_ports = [port.accept(self,env) 
                        for port in process_inst.actual_ports]
        port_bindings = map(lambda a, b: (a,b), 
                            processdef.ports, actual_ports)
        argument_values = [arg.accept(self, env) 
                            for arg in process_inst.actual_params]
        param_bindings = map(lambda a, b: (a,b), 
                             processdef.parameters, argument_values)
        new_bindings = dict(port_bindings + param_bindings)
        new_env = Env(Frame(new_bindings), latent_process.env)
        return processdef.body.accept(self, new_env, *args[1:], **kwargs)
    
    def visit_localdefs(self, localdefs, *args, **kwargs):
        log.note(2, "1 -- %s", localdefs)
        currentThread().check_life()
        env = args[0]
        new_env = Env(Frame(), env)
        for d in localdefs.defs:
            currentThread().check_life()
            d.accept(self, new_env)
        return localdefs.body.accept(self, new_env, *args[1:], **kwargs)
        
    def visit_indexedpar(self, indexedpar, *args, **kwargs):
        log.note(2, "1 -- %s", indexedpar)
        me = currentThread()
        me.check_life()
        env = args[0]
        process = indexedpar.process
        pattern = indexedpar.pattern
        sequence = indexedpar.sequence
        threads = []
        seq_obj = sequence.accept(self, env)
        if not isinstance(seq_obj, Tuple):
            raise RuntimeException(indexedpar, "type error: incorrect sequence for indexed-parallel")
        for item in seq_obj.value:
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                new_args = (self, new_env) + args[1:]
                threads += [KThread(target = process, args = new_args, kwargs = kwargs)]
        me.subthreads = threads
        if hasattr(me, 'signal'):
            for thread in threads:
                thread.signal = me.signal
        for thread in threads:
            thread.start()
        self.join(threads)
        
    def visit_loop(self, loop, *args, **kwargs):
        log.note(2, "1 -- %s", loop)
        currentThread().check_life()
        env = args[0]
        process = loop.process
        pattern = loop.pattern
        sequence = loop.sequence
        seq_obj = sequence.accept(self, env)
        if not isinstance(seq_obj, Tuple):
            raise RuntimeException(loop, "type error: incorrect sequence for indexed-sequence")
        for item in seq_obj.value:
            #~ currentThread().check_life()
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                value = process.accept(self, new_env, *args[1:], **kwargs)

    def visit_interrupt(self, interrupt, *args, **kwargs):
        log.note(2, "1 -- %s", interrupt)
        me = currentThread()
        me.check_life()
        log.note(2, "2 -- %s", interrupt)
        env = args[0]
        base_thread = KThread(target = interrupt.base_process, args = (self, env) + args[1:], kwargs = kwargs)
        signal = KEvent()
        base_thread.register_blocking_resource(signal)  # Unnecessary if base_thread blocks by calling KEvent.wait, signal will be added to the base_thread.__blocking_resources by the wait, but necessary since base_thread may finish before the timeout so it needs to wakeup signal when calling finish.
        alt_thread = KThread(target = interrupt.alt_process, args = (self, env, signal), kwargs = kwargs)
        alt_thread.signal = signal
        me.subthreads += [base_thread, alt_thread]
        log.note(2, "3 -- %s\n\n%s\n", interrupt, self.main_thread.tree_str())
        base_thread.start()
        alt_thread.start()
        signal.wait()
        log.note(2, "4 -- %s", interrupt)
        if not signal.isSet():
            if alt_thread.isAlive():
                # Base thread finished first
                log.note(2, "5 -- %s", interrupt)
                alt_thread.kill()
                log.note(2, "6 -- %s", interrupt)
                base_thread.remove_blocking_resource(signal)
                base_thread.recursive_join()
                return
        else:
            if base_thread.isAlive():
                # Alt thread interacted
                log.note(2, "7 -- %s", interrupt)
                base_thread.kill()
                log.note(2, "8 -- %s", interrupt)
                base_thread.remove_blocking_resource(signal)
            if hasattr(me, 'signal'):
                log.note(2, "9 -- %s", interrupt)
                me.signal.set()
            alt_thread.recursive_join()
            return
        log.note(2, "10 -- %s", interrupt)
        base_thread.remove_blocking_resource(signal)
            

        #~ class BaseController:
            #~ def __call__(self):
                #~ interrupt.base_process.accept(self, env, *args, **kwargs)
                #~ self.alt_controller_thread.kill()
        #~ class AltController:
            #~ def __call__(self):
                #~ alt_thread = KThread(target = interrupt.alt_process,
                #~ interrupt.alt_process.accept(self, env, signal, *args[1:], **kwargs)
                #~ signal.wait()
                #~ if signal.isSet() and base_thread.isAlive():
                    #~ base_thread.kill()
                    #~ self.base_controller.kill()
                    #~ base_controller_thread.kill()
        #~ base_controller = BaseController()
        #~ alt_controller = AltController()
        #~ base_controller_thread = KThread(target = base_controller)
        #~ alt_controller_thread = KThread(target = alt_controller)
        #~ base_controller.alt_controller_thread = alt_controller_thread
        #~ alt_controller.base_controller_thread = base_controller_thread
        #~ currentThread().subthreads = [base_thread, alt_thread, base_controller_thread, alt_controller_thread]
        #~ base_controller_thread.start()
        #~ alt_controller_thread.start()
        #~ self.join([base_controller_thread, alt_controller_thread])
    
    def visit_wait(self, wait, *args, **kwargs):
        log.note(2, "1 -- %s", wait)
        currentThread().check_life()
        env = args[0]
        try:
            delay_value = wait.delay.accept(self, env)
            #~ self.clock.sleep(float(delay_value))
            alarm = KEvent()
            alarm.wait(float(delay_value))
            log.note(2, "2 -- %s", wait)
            wait.continuation.accept(self, *args, **kwargs)
            log.note(2, "3 -- %s", wait)
        except TypeError:
            log.note(2, "4 -- %s", wait)
            raise RuntimeException(wait, "type error: wait argument is not a valid number")
    
    def visit_timeout(self, timeout, *args, **kwargs):
        log.note(2, "1 -- %s", timeout)
        me = currentThread()
        me.check_life()
        log.note(2, "2 -- %s", timeout)
        env = args[0]
        signal = KEvent()
        base_thread = KThread(target = timeout.base_process, args = (self, env) + args, kwargs = kwargs)
        base_thread.signal = signal
        me.subthreads += [base_thread]
        log.note(2, "3 -- %s\n%s", timeout, me.tree_str())
        base_thread.register_blocking_resource(signal)  # Unnecessary if base_thread blocks by calling KEvent.wait, signal will be added to the base_thread.__blocking_resources by the wait, but necessary since base_thread may finish before the timeout so it needs to wakeup signal when calling finish.
        time_limit = timeout.time_limit.accept(self, env)
        base_thread.start()
        try:
            signal.wait(float(time_limit))
            log.note(2, "4 -- %s", timeout)
            if not signal.isSet() and base_thread.isAlive():
                log.note(2, "5 -- %s", timeout)
                base_thread.kill() #Check: this wakes up signal; what if parent thread shares signal?
                log.note(2, "6 -- %s", timeout)
                base_thread.recursive_join()
                log.note(2, "7 -- %s", timeout)
                return timeout.alt_process.accept(self, *args)
            elif signal.isSet():
                log.note(2, "8 -- %s", timeout)
                if hasattr(me, 'signal'):
                    log.note(2, "9 -- %s", timeout)
                    me.signal.set()
            base_thread.recursive_join()
            log.note(2, "10 -- %s", timeout)
        except TypeError:
            log.note(2, "11 -- %s", timeout)
            base_thread.kill()
            log.note(2, "12 -- %s", timeout)
            raise RuntimeException(timeout, "type error: timeout argument is not a valid number")
    
    def visit_sendat(self, sendat, *args, **kwargs):
        log.note(2, "1 -- %s", sendat)
        me = currentThread()
        me.check_life()
        env = args[0]
        time_before_action = self.clock.time
        #~ sendat.action.path = sendat.get_full_location()
        sendat.action.accept(self, env)
        time_after_action = self.clock.time
        elapsed_time = Number(time_after_action - time_before_action)
        if hasattr(me, 'signal'):
            me.signal.set()
        new_frame = Frame({sendat.elapsed: elapsed_time})
        new_env = Env(new_frame, env)
        sendat.continuation.accept(self, new_env, *args[1:])
        
    def visit_receiveat(self, receiveat, *args, **kwargs):
        log.note(2, "1 -- %s", receiveat)
        me = currentThread()
        me.check_life()
        env = args[0]
        time_before_action = self.clock.time
        data = receiveat.action.accept(self, env)
        time_after_action = self.clock.time
        elapsed_time = Number(time_after_action - time_before_action)
        if hasattr(me, 'signal'):
            me.signal.set()
        new_frame = Frame({receiveat.parameter: data, receiveat.elapsed: elapsed_time})
        new_env = Env(new_frame, env)
        receiveat.continuation.accept(self, new_env, *args[1:])
    
    # Toplevel modules

    def visit_funcdef(self, funcdef, *args, **kwargs):
        env = args[0]
        env.add_binding(funcdef.name, LatentFunction(funcdef, env))

    def visit_processdef(self, processdef, *args, **kwargs):
        env = args[0]
        env.add_binding(processdef.name, LatentProcess(processdef, env))
    
    def visit_module(self, module, *args, **kwargs):
        log.note(2, "visit module -- %s" % module)
        env = args[0]
        new_env = Env(Frame(), env)
        for processdef in module.process_defs:
            processdef.accept(self, new_env)
        if 'main' in kwargs and kwargs['main'] is not None:
            main = kwargs['main']
        else:
            main = 'main'
        main_proc = new_env[main]
        main_ports = []
        if self.server is not None and len(main_proc.processdef.ports) > 0:
            port_to_server = main_proc.processdef.ports[0]
            server_link = RemoteAsyncChannel(self.socket)
            new_env.add_binding(port_to_server, server_link)
            main_ports.append(Var(port_to_server))
        main_inst = ProcessInstantiation(main, main_ports, [])
        main_inst.accept(self, new_env, **kwargs)
        #~ new_args = tuple([self, new_env] + list(args[1:]))
        #~ main_thread = KThread(target = main_inst, args = new_args, kwargs = kwargs)
        #~ main_thread.start()
        #~ main_thread.join()
        self.keyboard_handler.alive = False

    # Expression
    
    def visit_bottom(self, bottom, *args, **kwargs):
        return Bottom
    
    def visit_variable(self, var, *args, **kwargs):
        env = args[0]
        try:
            result = env[var.name]
            return result
        except EnvironmentException:
            raise RuntimeException(var.term, "incorrect variable access: variable '%s' is not defined in the current environment."%var.name)
    
    def visit_boolconst(self, const, *args, **kwargs):
        return Boolean(const.value)

    def visit_compare(self, comp, *args, **kwargs):
        env = args[0]
        lhs = comp.lhs.accept(self, env).value
        rhs = comp.rhs.accept(self, env).value
        rel = comp.relation
        return Boolean(rel(lhs, rhs))
    
    def visit_boolbinexpr(self, expr, *args, **kwargs):
        env = args[0]
        left = expr.left.accept(self, env).value
        right = expr.right.accept(self, env).value
        op = expr.operator
        return Boolean(op(left, right))

    def visit_boolunexpr(self, expr, *args, **kwargs):
        env = args[0]
        arg = expr.argument.accept(self, env).value
        op = expr.operator
        return Boolean(op(arg))
        
    def visit_arithconst(self, const, *args, **kwargs):
        return Number(const.value)
        
    def visit_arithbinexpr(self, expr, *args, **kwargs):
        env = args[0]
        try:
            left = expr.left.accept(self, env).value
            right = expr.right.accept(self, env).value
            op = expr.operator
            return Number(op(left, right))
        except TypeError, e:
            raise RuntimeException(expr.term, "type error:" + str(e))

    def visit_arithunexpr(self, expr, *args, **kwargs):
        env = args[0]
        try:
            arg = expr.argument.accept(self, env).value
            op = expr.operator
            return Number(op(arg))
        except TypeError, e:
            raise RuntimeException(expr.term, "type error:" + str(e))
        
    def visit_string(self, string, *args, **kwargs):
        return String(string.msg)

    def visit_tuple(self, tuple_expr, *args, **kwargs):
        env = args[0]
        seq = tuple(item.accept(self, env) for item in tuple_expr.sequence)
        return Tuple(seq)

    def visit_cons(self, cons_expr, *args, **kwargs):
        env = args[0]
        try:
            front_expr = cons_expr.sequence[:-1]
            last_expr  = cons_expr.sequence[-1]
            front = tuple(item.accept(self, env) for item in front_expr)
            rest_val = last_expr.accept(self, env)
            if isinstance(rest_val, Tuple) and rest_val.value == []:
                rest = tuple([])
            else:
                rest = tuple(rest_val.value)
            return Tuple(front + rest)
        except TypeError, e:
            raise RuntimeException(cons_expr.term, "type error: " + str(e))
    
    def visit_comprehension(self, comprehension, *args, **kwargs):
        env = args[0]
        expression = comprehension.expression
        pattern = comprehension.pattern
        sequence = comprehension.sequence
        condition = comprehension.condition
        outcome = []
        for item in sequence.accept(self, env).value:
            #~ success, subst = match(pattern, item.accept(self, env)) 
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                if condition.accept(self, new_env).value:
                    outcome += [expression.accept(self, new_env)]
        return Tuple(outcome)
    
    def visit_tupleaccess(self, tupleaccess_expr, *args, **kwargs):
        env = args[0]
        try:
            tuple_val = tupleaccess_expr.tuple_expr.accept(self, env).value
            index = tupleaccess_expr.index_expr.accept(self, env).value
            return tuple_val[index]
        except IndexError, e:
            raise RuntimeException(tupleaccess_expr.term, "incorrect tuple access: tuple index out of range")
    
    def visit_funccall(self, funccall, *args, **kwargs):
        env = args[0]
        latent_func = funccall.var.accept(self, env)
        argument_values = [arg.accept(self, env) 
                            for arg in funccall.args]
        if isinstance(latent_func, BuiltInFunction):
            pyargs = [arg.pyvalue() for arg in argument_values]
            result = latent_func(*pyargs)
            klt_res = PyValue(result).klt_value
            return klt_res
        elif isinstance(latent_func, LatentFunction):
            funcdef = latent_func.funcdef
            param_bindings = map(lambda a, b: (a,b), 
                                 funcdef.parameters, argument_values)
            new_bindings = dict(param_bindings)
            new_env = Env(Frame(new_bindings), latent_func.env)
            return funcdef.expression.accept(self, new_env)
        else:
            raise RuntimeException(funccall.term, "type error: %s is not a function" % funcall.var.name)
        
    
    def visit_newchan(self, newchan, *args, **kwargs):
        if newchan.channel_class is Async:
            chcls = 'au'
        elif newchan.channel_class is Sync:
            chcls = 'su'
        return Channel(chcls)

    def visit_case(self, case_expr, *args, **kwargs):
        currentThread().check_life()
        env = args[0]
        expr_val = case_expr.expression.accept(self, env)
        for pattern, expr in case_expr.cases:
            corresponds, subst = match(pattern, expr_val)
            if corresponds:
                new_env = Env(Frame(subst), env)
                return expr.accept(self, new_env, *args[1:], **kwargs)
        return Bottom()

    def visit_condexpr(self, cond, *args, **kwargs):
        currentThread().check_life()
        env = args[0]
        condition_value = cond.condition.accept(self, env).value
        if condition_value is True:
            return cond.action.accept(self, *args, **kwargs)
        else:
            return cond.alternative.accept(self, *args, **kwargs)

    def visit_letexpr(self, let, *args, **kwargs):
        currentThread().check_life()
        env = args[0]
        bindings = dict((var, let.bindings[var].accept(self, env)) 
                        for var in let.bindings)
        new_env = Env(Frame(bindings), env)
        return let.body.accept(self, new_env, *args[1:], **kwargs)

     


    # Other methods

    def join(self, threads):
        if threads is not None:
            for thread in threads:
                thread.join()
    
    def builtins(self):
        return setup_built_ins()

############################################################################
            


def match(pattern, datum, subst=None):
    if subst is None:
        subst = {}
    if isinstance(pattern, ConstantExpression):
        return pattern.value == datum.value, subst
    elif isinstance(pattern, Variable):
        if pattern.name in subst and subst[pattern.name] != datum:
            print "Mismatch: '" + pattern.name + "' was previously matched with '" + str(subst[pattern.name]) + "'"
            return False, {}
        subst[pattern.name] = datum
        return True, subst
    elif isinstance(pattern, TupleExpression):
        if not isinstance(datum, Tuple) \
        or len(pattern.sequence) != len(datum.value):
            return False, subst
        correspond, new_subst = True, subst
        i = 0
        while i < len(pattern.sequence) and correspond:
            correspond, new_subst = match(pattern.sequence[i], datum.value[i], new_subst)
            i += 1
        return correspond, new_subst
    elif isinstance(pattern, ConsExpression):
        if not isinstance(datum, Tuple) \
        or len(pattern.sequence) > len(datum.value) + 1:
            return False, subst
        correspond, new_subst = True, subst
        i = 0
        while i < len(pattern.sequence) - 1 and correspond:
            correspond, new_subst = match(pattern.sequence[i], datum.value[i], new_subst)
            i += 1
        if correspond:
            correspond, new_subst = match(pattern.sequence[i], Tuple(datum.value[i:]), new_subst)
        return correspond, new_subst
    return False, subst


class KeyboardInterruptHandler:
    
    def __init__(self, cycle_duration=1.0, target=None):
        self.alive = True
        self.cycle_duration = cycle_duration
        self.target = target
    
    def run(self):
        try:
            while self.alive:
                sleep(self.cycle_duration)
            log.note(1, "death by natural cause")
        except KeyboardInterrupt:
            print_time("Keyboard interrupt")
        if self.target is not None:
            log.note(1, "killing target -- %s" % self.target)
            #~ self.target.hardkill()
            self.target.kill()
        log.note(1, "death")
