"""
Project: kiltera 
Module: abstract_syntax
Author: Ernesto Posse
Created on: Feb 22, 2006
Last modified: Jan 22, 2007

Description:
  This module provides the abstract syntax of kiltera processes.
"""

from time import *
import bisect
from threading import currentThread
from posse_tools.utils import all, any, print_time, unzip


# Atoms

class Atom(object):
  
    def __init__(self):
        self.__parent = None
    
    def get_parent(self):
        return self.__parent
    
    def set_parent(self, parent):
        self.__parent = parent
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if isinstance(attr, Expression):
                attr.term = parent
    
    parent = property(get_parent, set_parent)
  
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_atom(self, *args, **kwargs)
        
    def __call__(self, visitor, *args):
        """This method is necessary for executing the process as a thread"""
        return self.accept(visitor, *args)
    
    


class Output(Atom):
    def __init__(self, channel_ref, data=None):
        Atom.__init__(self)
        self.channel_ref = channel_ref
        self.data = data
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_output(self, *args, **kwargs)


class Input(Atom):
    def __init__(self, channel_ref, parameter=''):
        Atom.__init__(self)
        self.channel_ref = channel_ref
        self.parameter = parameter
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_input(self, *args, **kwargs)


class Assign(Atom):
    def __init__(self, variable, expression):
        Atom.__init__(self)
        self.variable = variable
        self.expression = expression
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_assign(self, *args, **kwargs)


class Print(Atom):
    def __init__(self, msg):
        Atom.__init__(self)
        self.msg = msg
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_print(self, *args, **kwargs)

# Processes

class Process(object):
    
    def __init__(self, pos=None):
        self.parent = None
        self.__subterms = []
        self.__subinstances = []
        self.pos = pos
        self.__finished = False
        self.__interacted = False
        self.events = []
    
    def __repr__(self):
        if self.pos is None:
            pos = ''
        else:
            pos = ' at ' + str(self.pos[:2])
        return "%s%s" % (self.__class__.__name__, pos)
        
    def get_subterms(self):
        return self.__subterms 
        
    def set_subterms(self, proc_list):
        #~ print 'setting subterms of ',self,proc_list
        self.__subterms = proc_list
        for component in self.__subterms:
            component.parent = self
        self.threads = []
            
    def del_subterms(self):
        for component in self.__subterms:
            del component
        del self.__subterms
        
    subterms = property(get_subterms, set_subterms, del_subterms)
    
    def get_finished(self):
        if self.__finished:
            return True
        self.__finished = all(component.finished for component in self.__subterms)
        return self.__finished
    
    def set_finished(self, val):
        self.__finished = val
        
    def update_finished(self):
        """This method is meant to be invoked by the child processes"""
        self.__finished = all(component.finished for component in self.__subterms)
        if self.parent is not None:
            self.parent.update_finished()
    
    finished = property(get_finished, set_finished)
    
    def get_interacted(self):
        #~ print 1, self.__interacted, self
        if self.__interacted:
            #~ print 2, self.__interacted, self
            return True
        #~ print 3, self.__interacted, self
        self.__interacted = any(component.interacted for component in self.__subterms)
        #~ print 4, self.__interacted, self
        return self.__interacted
    
    def set_interacted(self, val):
        self.__interacted = val
    
    def update_interacted(self):
        """This method is meant to be invoked by the child processes"""
        self.__interacted = any(component.interacted for component in self.__subterms)
        if self.parent is not None:
            self.parent.update_interacted()
    
    interacted = property(get_interacted, set_interacted)
    
    def add_subinstance(self, instance):
        self.__subinstances.append(instance)
    
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_process(self, *args, **kwargs)
        
    def __call__(self, visitor, *args, **kwargs):
        """This method is necessary for executing the process as a thread"""
        return self.accept(visitor, *args, **kwargs)
    
    #~ def kill(self):
        #~ print "killing process",self,self.threads
        #~ for t in self.threads:
            #~ t.kill()
        #~ for c in self.__subterms:
            #~ c.kill()
    
    #~ def get_thread(self):
        #~ if self.parent is None:
            #~ return "Main thread"
        #~ else:
            #~ for t in self.parent.threads:
                #~ if t.target == self:
                    #~ return t
            #~ return self.parent.get_thread()
    
    def get_cenap(self):
        """Get the Closest Enclosing Named Ancestor Process"""
        if self.parent is None:
            return None
        elif isinstance(self.parent, ProcessDefinition):
            return self.parent
        return self.parent.get_cenap()
    
    def get_full_location(self):
        """Get the fully-qualified location, i.e. the list of enclosing 
            named processes in nesting order"""
        full_location = []
        named_ancestor = self.get_cenap()
        while named_ancestor is not None:
            full_location = [named_ancestor] + full_location
            named_ancestor = named_ancestor.get_cenap()
        return full_location
    
    def cancel(self, event_scheduler):
        """Remove all events associated with this process from the event list"""
        for ev in self.events:
            event_scheduler.cancel(ev)
        self.events = []
        for c in self.__subterms:
            c.cancel(event_scheduler)
    
    def add_event(self, event):
        bisect.insort(self.events, event)
    
    def remove_event(self, event):
        i = bisect.bisect(self.events, event)
        if i > 0:
            del self.events[i-1]


class Nil(Process):
    def __init__(self, pos=None):
        Process.__init__(self, pos)
        self.__finished = False
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_nil(self, *args, **kwargs)
    def get_finished(self):
        return self.__finished
    finished = property(get_finished, Process.set_finished)


class Done(Process):
    def __init__(self, pos=None):
        Process.__init__(self, pos)
        self.__finished = True
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_done(self, *args, **kwargs)
    def get_finished(self):
        return self.__finished
    finished = property(get_finished, Process.set_finished)


class Prefix(Process):
    def __init__(self, action, continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.action = action
        self.continuation = continuation
        self.set_subterms([self.continuation])
        self.action.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_prefix(self, *args, **kwargs)


class Send(Process):
    def __init__(self, channel_ref, data=None, continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.channel_ref = channel_ref
        self.continuation = continuation
        self.action = Output(channel_ref, data)
        self.set_subterms([self.continuation])
        self.action.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_send(self, *args, **kwargs)
    #~ def kill(self):
        #~ print "Killing send",self.pos,self,self.get_thread()
        #~ if hasattr(self.action,'channel'):
            #~ print self.action.channel
            #~ self.action.channel.wakeup()
        #~ print "..KS",self.threads,self.get_subterms()
        #~ for t in self.threads:
            #~ t.kill()
        #~ for c in self.get_subterms():
            #~ c.kill()
        #~ print "raising SystemExit",self.get_thread()
        #~ raise SystemExit()


class Receive(Process):
    def __init__(self, channel_ref, parameter='', continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.channel_ref = channel_ref
        self.parameter = parameter
        self.continuation = continuation
        self.action = Input(channel_ref, parameter)
        self.set_subterms([self.continuation])
        self.action.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_receive(self, *args, **kwargs)
    #~ def kill(self):
        #~ print "Killing receive",self.pos,self,self.get_thread()
        #~ if hasattr(self.action,'channel'):
            #~ print self.action.channel
            #~ print_time("killing receive 1 %s -- %s"%(currentThread(),self.action.channel))
            #~ self.action.channel.wakeup()
        #~ print "..KR",self.threads,self.get_subterms()
        #~ for t in self.threads:
            #~ t.kill()
        #~ for c in self.get_subterms():
            #~ c.kill()
        #~ print "raising SystemExit",self.get_thread()
        #~ raise SystemExit()


class Parallel(Process):
    def __init__(self, process_list, pos=None):
        Process.__init__(self, pos)
        self.process_list = list(process_list)
        self.set_subterms(self.process_list)
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_parallel(self, *args, **kwargs)


class Sequence(Process):
    def __init__(self, process_list, pos=None):
        Process.__init__(self, pos)
        self.process_list = list(process_list)
        self.set_subterms(self.process_list)
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_sequence(self, *args, **kwargs)


class Local(Process):
    def __init__(self, channel_class, channel_names, process, pos=None):
        Process.__init__(self, pos)
        self.channel_class = channel_class
        self.channel_names = channel_names
        self.process = process
        self.set_subterms([self.process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_local(self, *args, **kwargs)

class ChannelArray(Process):
    def __init__(self, channel_class, name, size, process, pos=None):
        Process.__init__(self, pos)
        self.channel_class = channel_class
        self.name = name
        self.size = size
        self.process = process
        self.set_subterms([self.process])
        self.size.term = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_chanarray(self, *args, **kwargs)


class Let(Process):
    def __init__(self, bindings, process, pos=None):
        Process.__init__(self, pos)
        self.bindings = bindings
        self.process = process
        self.set_subterms([self.process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_let(self, *args, **kwargs)


class Conditional(Process):
    def __init__(self, condition, action, alternative, pos=None):
        Process.__init__(self, pos)
        self.condition = condition
        self.action = action
        self.alternative = alternative
        self.set_subterms([self.action, self.alternative])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_conditional(self, *args, **kwargs)


class Match(Process):
    def __init__(self, expression, cases, pos=None):
        Process.__init__(self, pos)
        self.expression = expression
        self.cases = cases
        self.set_subterms([proc for pattern, proc in self.cases])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_match(self, *args, **kwargs)


class ProcessInstantiation(Process):
    def __init__(self, process_name, actual_ports, actual_params=None, pos=None):
        Process.__init__(self, pos)
        self.process_name = process_name
        self.actual_ports = actual_ports
        self.actual_params = actual_params
        if actual_params is None:
            self.actual_params = []
        self.set_subterms([])
        self.__finished = False
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_procinst(self, *args, **kwargs)
    def get_finished(self):
        if self.__finished:
            #~ print 2, self.__finished, self
            return True
        #~ print 3, self.__finished, self
        self.__finished = all(component.finished for component in self.__subterms)
        #~ print 4, self.__finished, self
        return self.__finished
    finished = property(get_finished, Process.set_finished)

class LocalDefinitions(Process):
    def __init__(self, defs, body, pos=None):
        Process.__init__(self, pos)
        self.defs = defs
        self.body = body
        self.set_subterms([self.body])
        for subdef in self.defs:
            subdef.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_localdefs(self, *args, **kwargs)

class IndexedParallel(Process):
    def __init__(self, process, pattern, sequence, pos=None):
        Process.__init__(self, pos)
        self.process = process
        self.pattern = pattern
        self.sequence = sequence
        self.set_subterms([self.process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_indexedpar(self, *args, **kwargs)

class Loop(Process):
    def __init__(self, process, pattern, sequence, pos=None):
        Process.__init__(self, pos)
        self.process = process
        self.pattern = pattern
        self.sequence = sequence
        self.set_subterms([self.process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_loop(self, *args, **kwargs)


class Interrupt(Process):
    def __init__(self, base_process, alt_process, pos=None):
        Process.__init__(self, pos)
        self.base_process = base_process
        self.alt_process = alt_process
        self.set_subterms([self.base_process, self.alt_process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_interrupt(self, *args, **kwargs)

class Wait(Process):
    def __init__(self, delay, continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.delay = delay
        self.continuation = continuation
        self.set_subterms([self.continuation])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_wait(self, *args, **kwargs)


class Timeout(Process):
    def __init__(self, base_process, time_limit, alt_process, pos=None):
        Process.__init__(self, pos)
        self.base_process = base_process
        self.time_limit = time_limit
        self.alt_process = alt_process
        self.set_subterms([self.base_process, self.alt_process])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_timeout(self, *args, **kwargs)


class SendAt(Process):
    def __init__(self, channel_ref, data=None, elapsed=None, continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.channel_ref = channel_ref
        self.continuation = continuation
        self.action = Output(channel_ref, data)
        self.elapsed = elapsed
        self.set_subterms([self.continuation])
        self.action.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_sendat(self, *args, **kwargs)


class ReceiveAt(Process):
    def __init__(self, channel_ref, parameter='', elapsed=None, continuation=Done(), pos=None):
        Process.__init__(self, pos)
        self.channel_ref = channel_ref
        self.parameter = parameter
        self.continuation = continuation
        self.action = Input(channel_ref, parameter)
        self.elapsed = elapsed
        self.set_subterms([self.continuation])
        self.action.parent = self
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_receiveat(self, *args, **kwargs)


# Module toplevel 

class FunctionDefinition:
    def __init__(self, name, parameters, expression, pos=None):
        self.name = name
        self.parameters = parameters
        self.expression = expression
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_funcdef(self, *args, **kwargs)

class ProcessDefinition(Process):
    def __init__(self, name, ports, parameters, body, pos=None):
        Process.__init__(self, pos)
        self.name = name
        self.ports = ports
        self.parameters = parameters
        self.body = body
        self.parent = None
        self.set_subterms([self.body])
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_processdef(self, *args, **kwargs)

class Module:
    def __init__(self, name, imports, process_defs):
        self.name = name
        self.imports = imports
        self.process_defs = process_defs
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_module(self, *args, **kwargs)
    def __call__(self, visitor, *args, **kwargs):
        """This method is necessary for executing the process as a thread"""
        return self.accept(visitor, *args, **kwargs)
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)

# Expressions

class Expression(object):
    
    def __init__(self, pos=None):
        self.pos = pos
        self.parent = None
        self.__subexpressions = []
        self.__term = None
    
    def get_subexpressions(self):
        return self.__subexpressions
    
    def set_subexpressions(self, subexpressions):
        self.__subexpressions = subexpressions
        for expression in self.__subexpressions:
            expression.parent = self
    
    subexpressions = property(get_subexpressions, set_subexpressions)

    def get_term(self):
        return self.__term
    
    def set_term(self, term):
        self.__term = term
        for expression in self.__subexpressions:
            expression.term = term
    
    term = property(get_term, set_term)
    
    def seq(self):
        return [self]


class ConstantExpression(Expression):
    def __init__(self,pos=None):
        Expression.__init__(self,pos)


class BottomExpr(ConstantExpression):
    def __init__(self, pos=None):
        ConstantExpression.__init__(self,pos)
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_bottom(self, *args, **kwargs)
    
    
class VariableAccess(Expression):
    def __init__(self, name, pos=None):
        Expression.__init__(self,pos)
        self.name = name
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_variable(self, *args, **kwargs)
    def __str__(self):
        return self.name
    def __repr__(self):
        return self.name


class BooleanExpression(Expression):
    def __init__(self, pos=None):
        Expression.__init__(self,pos)


class BooleanConstant(BooleanExpression, ConstantExpression):
    def __init__(self, value, pos=None):
        BooleanExpression.__init__(self,pos)
        self.value = value
        self.operator = BooleanConstant
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_boolconst(self, *args, **kwargs)


class Comparison(BooleanExpression):
    def __init__(self, relation, lhs, rhs, pos=None):
        BooleanExpression.__init__(self,pos)
        self.relation = relation
        self.lhs = lhs
        self.rhs = rhs
        self.operator = relation
        self.subexpressions = [self.lhs, self.rhs]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_compare(self, *args, **kwargs)


class BooleanBinaryExpression(BooleanExpression):
    def __init__(self, operator, left, right, pos=None):
        BooleanExpression.__init__(self,pos)
        self.operator = operator
        self.left = left
        self.right = right
        self.subexpressions = [self.left, self.right]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_boolbinexpr(self, *args, **kwargs)


class BooleanUnaryExpression(BooleanExpression):
    def __init__(self, operator, argument, pos=None):
        BooleanExpression.__init__(self,pos)
        self.operator = operator
        self.argument = argument
        self.subexpressions = [self.argument]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_boolunexpr(self, *args, **kwargs)


class ArithmeticExpression(Expression):
    def __init__(self, pos=None):
        Expression.__init__(self,pos)


class NumericConstant(ArithmeticExpression, ConstantExpression):
    def __init__(self, value, pos=None):
        ConstantExpression.__init__(self,pos)
        self.value = value
        self.operator = NumericConstant
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_arithconst(self, *args, **kwargs)


class ArithmeticBinaryExpression(ArithmeticExpression):
    def __init__(self, operator, left, right, pos=None):
        ArithmeticExpression.__init__(self,pos)
        self.operator = operator
        self.left = left
        self.right = right
        self.subexpressions = [self.left, self.right]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_arithbinexpr(self, *args, **kwargs)


class ArithmeticUnaryExpression(ArithmeticExpression):
    def __init__(self, operator, argument, pos=None):
        ArithmeticExpression.__init__(self,pos)
        self.operator = operator
        self.argument = argument
        self.subexpressions = [self.argument]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_arithunexpr(self, *args, **kwargs)


class StringConstant(ConstantExpression):
    def __init__(self, msg, pos=None):
        ConstantExpression.__init__(self,pos)
        self.msg = msg
        self.value = msg
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_string(self, *args, **kwargs)


class TupleExpression(Expression):
    def __init__(self, expr_seq, pos=None):
        Expression.__init__(self,pos)
        self.sequence = expr_seq
        self.subexpressions = self.sequence
    def seq(self):
        return self.sequence
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_tuple(self, *args, **kwargs)

class ConsExpression(Expression):
    def __init__(self, expr_seq, pos=None):
        Expression.__init__(self,pos)
        self.sequence = expr_seq
        self.subexpressions = self.sequence
    def seq(self):
        return self.sequence
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_cons(self, *args, **kwargs)

class Comprehension(Expression):
    def __init__(self, expression, pattern, sequence, condition, pos=None):
        Expression.__init__(self,pos)
        self.expression = expression
        self.pattern = pattern
        self.sequence = sequence
        self.condition = condition
        self.subexpressions = [self.expression, self.pattern, self.sequence, self.condition]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_comprehension(self, *args, **kwargs)

class TupleAccess(Expression):
    def __init__(self, tuple_expr, index_expr, pos=None):
        Expression.__init__(self,pos)
        self.tuple_expr = tuple_expr
        self.index_expr = index_expr
        self.subexpressions = [self.tuple_expr, self.index_expr]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_tupleaccess(self, *args, **kwargs)

class FunctionCall(Expression):
    def __init__(self, var, args, pos=None):
        Expression.__init__(self,pos)
        self.var = var
        self.args = args
        self.subexpressions = [self.var] + self.args
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_funccall(self, *args, **kwargs)

class NewChannel(ConstantExpression):
    def __init__(self, channel_class, pos=None):
        ConstantExpression.__init__(self,pos)
        self.channel_class = channel_class
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_newchan(self, *args, **kwargs)

class CaseExpression(Expression):
    def __init__(self, expression, cases, pos=None):
        Expression.__init__(self,pos)
        self.expression = expression
        self.cases = cases
        self.subexpressions = reduce(lambda x,y: x+y, unzip([(pattern, expr) for pattern, expr in self.cases]))
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_case(self, *args, **kwargs)
        
class CondExpression(Expression):
    def __init__(self, condition, action, alternative, pos=None):
        Expression.__init__(self, pos)
        self.condition = condition
        self.action = action
        self.alternative = alternative
        self.subexpressions = [self.condition, self.action, self.alternative]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_condexpr(self, *args, **kwargs)
    
class LetExpression(Expression):
    def __init__(self, bindings, process, pos=None):
        Expression.__init__(self, pos)
        self.bindings = bindings
        self.body = body
        self.subexpressions = [self.body]
    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_letexpr(self, *args, **kwargs)


# Synonyms

# ...for processes

Stop   = Nil
Pre    = Prefix
Action = Prefix
Par    = Parallel
Seq    = Sequence
Hide   = Local
New    = Local
FuncDef     = FunctionDefinition
ProcessDef  = ProcessDefinition
ProcessInst = ProcessInstantiation

# ...for expressions

Var      = VariableAccess
Variable = Var

BoolConst   = BooleanConstant
Compare     = Comparison
BoolExpr    = BooleanExpression
BoolBinExpr = BooleanBinaryExpression
BoolUnExpr  = BooleanUnaryExpression

NumConst     = NumericConstant
ArithExpr    = ArithmeticExpression
ArithBinExpr = ArithmeticBinaryExpression
ArithUnExpr  = ArithmeticUnaryExpression

StrConst  = StringConstant

TupleExpr = TupleExpression
TupleAcc  = TupleAccess


# Operators and relations for expressions

class BinaryOperator:
    def __init__(self, string, func):
        self.string = string
        self.func = func
    def __str__(self):
        return str(self.string)
    def __call__(self, left, right):
        return self.func(left, right)


class UnaryOperator:
    def __init__(self, string, func):
        self.string = string
        self.func = func
    def __str__(self):
        return str(self.string)
    def __call__(self, arg):
        return self.func(arg)


BinOp   = BinaryOperator
UnaryOp = UnaryOperator

Relation      = BinaryOperator
Rel           = Relation


Eq = Equals         = Rel('=',  lambda a,b: a == b)
Ne = Different      = Rel('!=', lambda a,b: a != b)
Lt = LessThan       = Rel('<',  lambda a,b: a < b)
Gt = GreaterThan    = Rel('>',  lambda a,b: a > b)
Le = LessOrEqual    = Rel('<=', lambda a,b: a <= b)
Ge = GreaterOrEqual = Rel('>=', lambda a,b: a >= b)

And = BinOp('and', lambda a,b: a and b)
Or  = BinOp('or', lambda a,b: a or b)
Not = UnaryOp('not ', lambda a: not a)

Add       = Plus   = BinOp('+', lambda a,b: a + b)
Substract = Minus  = BinOp('-', lambda a,b: a - b)
Multiply  = Times  = BinOp('*', lambda a,b: a * b)
Divide    = DivBy  = BinOp('/', lambda a,b: a / b)
Remainder = Mod    = BinOp('%', lambda a,b: a % b)
Negative  = UMinus = UnaryOp('-', lambda a: -a)

precedence = \
    {
        NumConst: 0, BoolConst: 0,
        Negative: 2, Not: 2,
        Multiply: 3, Divide: 3, Remainder: 3,
        Add: 4, Substract: 4,
        Lt: 5, Le: 5, Gt: 5, Ge: 5,
        Eq: 6, Ne: 6,
        And: 7,
        Or: 8,
    }


# Channel classes

Async = 0
Sync  = 1

