"""
Project: kiltera 
Module: ast_printer
Author: Ernesto Posse
Created on: Aug 11, 2006
Last modified: Jan 22, 2007

Description:
  This module provides the interface for visitors to use the visitor design pattern.
"""

from visitor import *
from kiltera_ast import *
from posse_tools.utils import *
#~ import kiltera_cfg.kiltera
#~ kiltera = kiltera_cfg.kiltera

class ASTPrinter(Visitor):
    
    def __init__(self, indent=2):
        self.indent = indent
    
    def pprint(self, term):
        print self.pformat(term)
    
    def pformat(self, term):
        return term.accept(self)
        
    # Atom visitors
    
    def visit_atom(self, atom, *args, **kwargs):
        return 'Atom' + '('+str(atom)+')'

    def visit_output(self, output, *args, **kwargs):
        ch_ref = output.channel_ref.accept(self)
        return 'Output' + '(' + output.data.accept(self) \
            + ', ' + ch_ref + ')'

    def visit_input(self, input, *args, **kwargs):
        ch_ref = input.channel_ref.accept(self)
        return 'Input' + '(' + "'" + str(input.parameter) + "'" \
            + ', ' + ch_ref + ')'

    def visit_assign(self, assignment, *args, **kwargs):
        lhs = str(assignment.variable.value)
        expr_str = assignment.expression.accept(self)
        return 'Assign' + '(' + str(lhs) + ', ' + expr_str + ')'

    def visit_print(self, print_atom):
        expr = print_atom.msg.accept(self)
        return 'Print' + '(' + expr + ')'
    
    # Process visitors
    
    def visit_process(self, process, *args, **kwargs):
        return 'Process' + '(' + str(process) + ')'

    def visit_nil(self, nil_expr, *args, **kwargs):
        return 'Nil' + '(' + ')'

    def visit_done(self, nil, *args, **kwargs):
        return 'Done' + '(' + ')'

    def visit_prefix(self, prefix, *args, **kwargs):
        result = 'Prefix' + '(' + '\n' + indent(prefix.action.accept(self), self.indent)
        result += ', \n' + \
            indent(prefix.continuation.accept(self), self.indent)
        return result + '\n)\n'

    def visit_send(self, send_proc, *args, **kwargs):
        return 'Send' + '(' + '\n' \
            + indent(send_proc.action.accept(self), self.indent) + ', \n' \
            + indent(send_proc.continuation.accept(self), self.indent) + '\n)\n'

    def visit_receive(self, receive_proc, *args, **kwargs):
        return 'Receive' + '(' + '\n' \
            + indent(receive_proc.action.accept(self), self.indent) + ', \n' \
            + indent(receive_proc.continuation.accept(self), self.indent) + '\n)\n'

    def visit_parallel(self, parallel_comp, *args, **kwargs):
        intro = 'Parallel'
        separator = ', \n'
        process_list = [indent(process.accept(self), self.indent) \
                        for process in parallel_comp.process_list]
        processes = separator.join(process_list)
        return intro + '(\n' + processes + '\n)\n'

    def visit_sequence(self, sequence_comp, *args, **kwargs):
        intro = 'Sequence'
        separator = ', \n'
        process_list = [indent(process.accept(self), self.indent) \
                        for process in sequence_comp.process_list]
        processes = separator.join(process_list)
        return intro + '(\n' + processes + '\n)\n'

    def visit_local(self, local, *args, **kwargs):
        intro = 'Channel'
        if local.channel_class is Async:
            kind = 'Async'
        elif local.channel_class is Sync:
            kind = 'Sync'
        separator = ', '
        names = separator.join("'" + str(name) +"'" for name in local.channel_names)
        return intro + '(' + kind + ', [' + names + '], \n' \
            + indent(local.process.accept(self), self.indent) + '\n)\n'
    
    def visit_chanarray(self, chanarray, *args, **kwargs):
        size = chanarray.size.accept(self)
        intro = 'ChannelArray'
        if chanarray.channel_class is Async:
            kind = 'Async'
        elif chanarray.channel_class is Sync:
            kind = 'Sync'
        return intro + '(' + kind + ', \'' + chanarray.name + '\', ' \
            + size + ', \n' \
            + indent(chanarray.process.accept(self), self.indent) + '\n)\n'

    def visit_let(self, let, *args, **kwargs):
        intro = 'Let'
        separator = '\n, '
        binding_list = ['(\'' + str(name) + '\', ' + str(let.bindings[name].accept(self)) + ')' \
                        for name in let.bindings]
        bindings = separator.join(binding_list)
        return intro + '(' + '[' + bindings + '], \n' \
            + indent(let.process.accept(self), self.indent) + '\n)\n'

    def visit_conditional(self, cond, *args, **kwargs):
        intro = 'Conditional'
        cond_text = cond.condition.accept(self)
        act_text = indent(cond.action.accept(self), self.indent)
        alt_text = indent(cond.alternative.accept(self), self.indent)
        return intro + '(' + cond_text + ',\n' \
            + act_text + ', \n' + alt_text + '\n)\n'
    
    def visit_match(self, match_proc, *args, **kwargs):
        intro = 'Match'
        separator = ',\n' 
        expr_text = match_proc.expression.accept(self)
        cases = []
        for pattern, process in match_proc.cases:
            cases += ['(' + pattern.accept(self) + \
                        ', \n' + \
                        indent(process.accept(self), self.indent) + '\n)']
        cases_text = separator.join(cases)
        return intro + '(' + expr_text + ',\n' \
            + indent('[\n' + indent(cases_text, self.indent) + '\n]\n', self.indent) \
            + '\n)\n'
    
    def visit_procinst(self, process_inst, *args, **kwargs):
        intro = 'ProcessInstantiation'
        separator = ', '
        actual_ports = [port.accept(self) for port in process_inst.actual_ports]
        ports = separator.join(actual_ports)
        if len(process_inst.actual_params) == 0:
            params = '()'
        else:
            arg_vals = [arg.accept(self) for arg in process_inst.actual_params]
            params = '(' + separator.join(arg_vals) + ')'
        return intro + '(' + "'" + str(process_inst.process_name) + "', " \
            + '[' + ports + '], ' + params + ')'
    
    def visit_localdefs(self, localdefs, *args, **kwargs):
        intro = 'LocalDefinitions'
        separator = ', \n'
        locals = separator.join([d.accept(self).strip() for d in localdefs.defs])
        body = localdefs.body.accept(self)
        return intro + '(\n' \
            +  indent('[\n' \
                      +  indent(locals, self.indent) \
                      + '\n],', \
                      self.indent) + '\n' \
            +  indent(body, self.indent) \
            + '\n)\n'
      
    def visit_indexedpar(self, indexedpar, *args, **kwargs):
        intro = 'IndexedParallel(\n'
        proc = indent(indexedpar.process.accept(self), self.indent)
        patt = indent(indexedpar.pattern.accept(self), self.indent)
        seq = indent(indexedpar.sequence.accept(self), self.indent)
        return intro + proc + ',\n' + patt + ',\n' + seq + '\n)\n'
    
    def visit_loop(self, loop, *args, **kwargs):
        intro = 'IndexedSequence(\n'
        proc = indent(loop.process.accept(self), self.indent) 
        patt = indent(loop.pattern.accept(self), self.indent)
        seq = indent(loop.sequence.accept(self), self.indent)
        return intro + proc + ',\n' + patt + ',\n' + seq + '\n)\n'

    def visit_interrupt(self, interrupt, *args, **kwargs):
        intro = 'Interrupt('
        base = interrupt.base_process.accept(self)
        alt  = interrupt.alt_process.accept(self)
        return intro + '\n' \
            + indent(base, self.indent) + ',\n' \
            + indent(alt, self.indent) + '\n)\n'

    def visit_wait(self, wait, *args, **kwargs):
        intro = 'Wait('
        delay_expr = wait.delay.accept(self)
        return intro + delay_expr + ', \n' \
            + indent(wait.continuation.accept(self), self.indent) + ')\n'
    
    def visit_timeout(self, timeout, *args, **kwargs):
        intro = 'Timeout('
        base = timeout.base_process.accept(self)
        expr = timeout.time_limit.accept(self)
        alt  = timeout.alt_process.accept(self)
        return intro + '\n' \
            + indent(base, self.indent) + ',\n' \
            + indent(expr, self.indent) + ',\n' \
            + indent(alt, self.indent) + '\n)\n'
    
    def visit_sendat(self, sendat, *args, **kwargs):
        return 'SendAt' + '(' + '\n' \
            + indent(sendat.action.accept(self), self.indent) + ', \n' \
            + indent("'" + str(sendat.elapsed) + "'", self.indent) + ', \n' \
            + indent(sendat.continuation.accept(self), self.indent) + '\n)\n'
        
    def visit_receiveat(self, receiveat, *args, **kwargs):
        return 'ReceiveAt' + '(' + '\n' \
            + indent(receiveat.action.accept(self), self.indent) + ', \n' \
            + indent("'" + str(receiveat.elapsed) + "'", self.indent) + ', \n' \
            + indent(receiveat.continuation.accept(self), self.indent) + '\n)\n'
    
    # Toplevel modules
    
    def visit_funcdef(self, funcdef, *args, **kwargs):
        intro = 'FunctionDefinition('
        separator = ', '
        param_list = separator.join(["'" + p + "'" for p in funcdef.parameters])
        param_list = '(' + param_list + ')'
        funcdef_text = intro + '\'' + str(funcdef.name) + '\', [' \
            + param_list + '],\n' \
            + indent(funcdef.expression.accept(self), self.indent) + '\n)\n'
        return funcdef_text
    
    def visit_processdef(self, processdef, *args, **kwargs):
        intro = 'ProcessDefinition('
        separator = ', '
        port_list = separator.join(["'" + p + "'" for p in processdef.ports])
        param_list = separator.join(["'" + p + "'" for p in processdef.parameters])
        param_list = '(' + param_list + ')'
        processdef_text = intro + '\'' + str(processdef.name) + '\', ' \
            + '[' + port_list + '], ' \
            + param_list + ',\n' \
            + indent(processdef.body.accept(self), self.indent) + '\n)\n'
        return processdef_text
        

    def visit_module(self, module, *args, **kwargs):
        intro = 'Module('
        separator = ',\n'
        imps = ', '.join([str(name) for name in module.imports])
        defs = [process_def.accept(self) for process_def in module.process_defs]
        module_text = intro + '\'' + str(module.name) + '\',\n'
        module_text += indent('['+separator.join(imps)+']', self.indent) + ',\n'
        module_text += indent('[\n' + indent(separator.join(defs), self.indent) + '\n]\n', self.indent) + ',\n'
        module_text += '\n)\n'
        return module_text

    # Expressions
    
    def visit_bottom(self, bottom, *args, **kwargs):
        return 'Bottom'
    
    def visit_variable(self, var, *args, **kwargs):
        return 'Var(\'' + str(var.name) + '\')'
    
    def visit_boolconst(self, const, *args, **kwargs):
        if bool(const.value):
            return 'BoolConst(True)'
        return 'BoolConst(False)'
    
    def visit_compare(self, comp, *args, **kwargs):
        lhs = comp.lhs.accept(self)
        rhs = comp.rhs.accept(self)
        rel = str(comp.relation)
        return lhs + ' ' + rel + ' ' + rhs
    
    def visit_boolbinexpr(self, expr, *args, **kwargs):
        left = expr.left.accept(self)
        right = expr.right.accept(self)
        op = str(expr.operator)
        if isinstance(expr.left, BoolExpr) \
        and precedence[expr.left.operator] > precedence[expr.operator]:
            left = '(' + left + ')'
        if isinstance(expr.right, BoolExpr) \
        and precedence[expr.right.operator] > precedence[expr.operator]:
            right = '(' + right + ')'
        return left + ' ' + op + ' ' + right
    
    def visit_boolunexpr(self, expr, *args, **kwargs):
        arg = expr.argument.accept(self)
        op = str(expr.operator)
        if isinstance(expr.argument, BoolExpr) \
        and precedence[expr.argument.operator] > precedence[expr.operator]:
            arg = '(' + arg + ')'
        return op + arg
        
    def visit_arithconst(self, const, *args, **kwargs):
        return 'Number(' + str(const.value) + ')'

    def visit_arithbinexpr(self, expr, *args, **kwargs):
        left = expr.left.accept(self)
        right = expr.right.accept(self)
        op = str(expr.operator)
        if isinstance(expr.left, ArithExpr) \
        and precedence[expr.left.operator] > precedence[expr.operator]:
            left = '(' + left + ')'
        if isinstance(expr.right, ArithExpr) \
        and precedence[expr.right.operator] > precedence[expr.operator]:
            right = '(' + right + ')'
        return left + ' ' + op + ' ' + right
    
    def visit_arithunexpr(self, expr, *args, **kwargs):
        arg = expr.argument.accept(self)
        op = str(expr.operator)
        if isinstance(expr.argument, ArithExpr) \
        and precedence[expr.argument.operator] > precedence[expr.operator]:
            arg = '(' + arg + ')'
        return op + arg 
    
    def visit_string(self, string, *args, **kwargs):
        return 'String("' + str(string.msg) + '")'

    def visit_tuple(self, tuple_expr, *args, **kwargs):
        separator = ', '
        items = [str(item.accept(self)) for item in tuple_expr.sequence]
        text = separator.join(items)
        if len(items) == 1:
            text += ','
        return 'TupleExpression(' + text + ')'

    def visit_cons(self, cons_expr, *args, **kwargs):
        separator = ', '
        items = [str(item.accept(self)) for item in cons_expr.sequence]
        text = separator.join(items)
        if len(items) == 1:
            text += ','
        return 'ConsExpression(' + text + ')'


    def visit_comprehension(self, comprehension, *args, **kwargs):
        intro = 'Comprehension('
        text = comprehension.expression.accept(self)
        text += ', '
        text += comprehension.pattern.accept(self)
        text += ', '
        text += comprehension.sequence.accept(self)
        text += ', '
        text += comprehension.condition.accept(self)
        return intro + text + ')'
    
    def visit_tupleaccess(self, tupleaccess_expr, *args, **kwargs):
        intro = 'TupleAccess'
        tuple_expr = tupleaccess_expr.tuple_expr.accept(self)
        index_expr = tupleaccess_expr.index_expr.accept(self)
        return intro + '(' + tuple_expr + ', ' + index_expr + ')'
    
    def visit_funccall(self, funccall, *args, **kwargs):
        intro = 'FunctionCall'
        separator = ', '
        arg_vals = [arg.accept(self) for arg in funccall.args]
        actual_args = '[' + separator.join(arg_vals) + ']'
        return intro + '(' + funccall.var.accept(self) + ',' + actual_args + ')'
    
    def visit_newchan(self, newchan, *args, **kwargs):
        intro = 'NewChannel'
        if newchan.channel_class is Async:
            chcls = 'Async'
        elif newchan.channel_class is Sync:
            chcls = 'Sync'
        return intro + '(' + chcls + ')'

    def visit_case(self, case_expr, *args, **kwargs):
        intro = 'CaseExpression'
        separator = ',\n' 
        expr_text = case_expr.expression.accept(self)
        cases = []
        for pattern, process in case_expr.cases:
            cases += ['(' + pattern.accept(self) + \
                        ', \n' + \
                        indent(process.accept(self), self.indent) + '\n)']
        cases_text = separator.join(cases)
        return intro + '(' + expr_text + ',\n' \
            + indent('[\n' + indent(cases_text, self.indent) + '\n]\n', self.indent) \
            + '\n)\n'

    def visit_condexpr(self, cond, *args, **kwargs):
        intro = 'CondExpression'
        cond_text = cond.condition.accept(self)
        act_text = indent(cond.action.accept(self), self.indent)
        alt_text = indent(cond.alternative.accept(self), self.indent)
        return intro + '(' + cond_text + ',\n' \
            + act_text + ', \n' + alt_text + '\n)\n'

    def visit_letexpr(self, let, *args, **kwargs):
        intro = 'LetExpression'
        separator = '\n, '
        binding_list = ['(\'' + str(name) + '\', ' + str(let.bindings[name].accept(self)) + ')' \
                        for name in let.bindings]
        bindings = separator.join(binding_list)
        return intro + '(' + '[' + bindings + '], \n' \
            + indent(let.process.accept(self), self.indent) + '\n)\n'

 