"""
Project: kiltera_syn 
Module: pretty_printer
Author: Ernesto Posse
Created on: Jun 9, 2006
Last modified: Jan 22, 2007

Description:
  This module provides a pretty printer for kiltera_syn.
"""

from visitor import *
from kiltera_ast import *
from posse_tools.utils import *
import kiltera_syn_cfg.kiltera_syn
kiltera_syn = kiltera_syn_cfg.kiltera_syn

class PrettyPrinter(Visitor):
    
    def __init__(self, indent=2):
        self.indent = indent
    
    def pprint(self, term):
        print self.pformat(term)
    
    def pformat(self, term):
        return term.accept(self)
        
    # Atom visitors
    
    def visit_atom(self, atom, *args, **kwargs):
        return str(atom)

    def visit_output(self, output, *args, **kwargs):
        ch_ref = output.channel_ref.accept(self)
        return kiltera_syn.send.string + ' ' + output.data.accept(self) \
            + ' ' + kiltera_syn.to.string + ' ' + ch_ref 

    def visit_input(self, input, *args, **kwargs):
        ch_ref = input.channel_ref.accept(self)
        return kiltera_syn.receive.string + ' ' + str(input.parameter) \
            + ' ' + kiltera_syn.fromkwd.string + ' ' + ch_ref

    def visit_assign(self, assignment, *args, **kwargs):
        lhs = str(assignment.variable.value)
        expr_str = assignment.expression.accept(self)
        return str(lhs) + ' ' + kiltera_syn.assign.string + ' ' + expr_str

    def visit_print(self, print_atom):
        expr = print_atom.msg.accept(self)
        return kiltera_syn.printkwd.string + ' ' + expr
    
    # Process visitors
    
    def visit_process(self, process, *args, **kwargs):
        return str(process)

    def visit_nil(self, nil_expr, *args, **kwargs):
        return kiltera_syn.nil.string + '\n'

    def visit_done(self, nil, *args, **kwargs):
        return kiltera_syn.done.string + '\n'

    def visit_prefix(self, prefix, *args, **kwargs):
        result = prefix.action.accept(self)
        if not isinstance(prefix.continuation, Nil):
            result += ' ' + kiltera_syn.arrow.string + ' \n' + \
                indent(prefix.continuation.accept(self), self.indent)
        return result + '\n'

    def visit_send(self, send_proc, *args, **kwargs):
        return send_proc.action.accept(self) + ' ' + kiltera_syn.arrow.string + ' \n' \
            + indent(send_proc.continuation.accept(self), self.indent) + '\n'

    def visit_receive(self, receive_proc, *args, **kwargs):
        return receive_proc.action.accept(self) + ' ' + kiltera_syn.arrow.string + ' \n' \
            + indent(receive_proc.continuation.accept(self), self.indent) + '\n'

    def visit_parallel(self, parallel_comp, *args, **kwargs):
        intro = kiltera_syn.parallel.string + '\n'
        #~ separator = kiltera_syn.parbars.string + ' \n'
        separator = ' \n'
        process_list = [indent(process.accept(self), self.indent) \
                        for process in parallel_comp.process_list]
        processes = separator.join(process_list)
        return intro + processes + '\n'

    def visit_sequence(self, sequence_comp, *args, **kwargs):
        intro = kiltera_syn.sequence.string + '\n'
        #~ separator = kiltera_syn.semicolon.string + ' \n'
        separator = ' \n'
        #~ separator = ' \n'
        process_list = [indent(process.accept(self), self.indent) \
                        for process in sequence_comp.process_list]
        processes = separator.join(process_list)
        return intro + processes +'\n'

    def visit_local(self, local, *args, **kwargs):
        intro = ''
        if local.channel_class is Async:
            intro = kiltera_syn.async.string
        elif local.channel_class is Sync:
            intro = kiltera_syn.sync.string
        intro += ' ' + kiltera_syn.channelkwd.string + ' '
        separator = ', '
        names = separator.join(str(name) for name in local.channel_names)
        return intro + names + ' ' + kiltera_syn.inkwd.string + ' \n' \
            + indent(local.process.accept(self), self.indent) + '\n'
    
    def visit_chanarray(self, chanarray, *args, **kwargs):
        size = chanarray.size.accept(self)
        intro = ''
        if chanarray.channel_class is Async:
            intro = kiltera_syn.async.string
        elif chanarray.channel_class is Sync:
            intro = kiltera_syn.sync.string
        intro += ' ' + kiltera_syn.channelskwd.string 
        return intro + ' ' + chanarray.name + ' ' \
            + kiltera_syn.lsqbra.string + size + kiltera_syn.rsqbra.string + ' ' \
            + kiltera_syn.inkwd.string + '\n'\
            + indent(chanarray.process.accept(self), self.indent) + '\n'

    def visit_let(self, let, *args, **kwargs):
        intro = kiltera_syn.let.string
        separator = '\n' + kiltera_syn.andop.string + ' '
        binding_list = [str(name) + ' = ' + str(let.bindings[name].accept(self)) \
                        for name in let.bindings]
        bindings = separator.join(binding_list)
        return intro + ' ' + bindings + '\n' + kiltera_syn.inkwd.string + '\n' \
            + indent(let.process.accept(self), self.indent) + '\n'

    def visit_conditional(self, cond, *args, **kwargs):
        intro = kiltera_syn.ifkwd.string
        cond_text = cond.condition.accept(self)
        act_text = indent(cond.action.accept(self), self.indent)
        alt_text = indent(cond.alternative.accept(self), self.indent)
        return intro + ' ' + cond_text + ' ' + kiltera_syn.then.string + ' \n' \
            + act_text + '\n' + kiltera_syn.elsekwd.string + '\n' + alt_text + '\n'
    
    def visit_match(self, match_proc, *args, **kwargs):
        intro = kiltera_syn.match.string
        separator = '\n' + kiltera_syn.bar.string + ' '
        then = ' ' + kiltera_syn.arrow.string + ' \n'
        expr_text = match_proc.expression.accept(self)
        cases = []
        for pattern, process in match_proc.cases:
            cases += [pattern.accept(self) + \
                        then + \
                        indent(process.accept(self), 2 + self.indent)]
        cases_text = separator.join(cases)
        return intro + ' ' + expr_text + ' ' + kiltera_syn.with.string + '\n' + ' ' * self.indent + cases_text
    
    def visit_procinst(self, process_inst, *args, **kwargs):
        separator = ', '
        actual_ports = [port.accept(self) for port in process_inst.actual_ports]
        ports = separator.join(actual_ports)
        if len(process_inst.actual_params) == 0:
            params = ''
        else:
            arg_vals = [arg.accept(self) for arg in process_inst.actual_params]
            params = kiltera_syn.lpar.string + separator.join(arg_vals) + kiltera_syn.rpar.string
        return str(process_inst.process_name) \
            + kiltera_syn.lsqbra.string + ports + kiltera_syn.rsqbra.string + params
    
    def visit_localdefs(self, localdefs, *args, **kwargs):
        body = localdefs.body
        intro = ''
        separator = ''
        locals = [d.accept(self) for d in localdefs.defs]
        locals_text = separator.join(locals)
        text = intro \
            + locals_text + '' \
            + kiltera_syn.inkwd.string + '\n' \
            + indent(body.accept(self), self.indent) + '\n'
        return text
        
    def visit_indexedpar(self, indexedpar, *args, **kwargs):
        intro = kiltera_syn.parallel.string + '\n'
        proc = indent(indexedpar.process.accept(self), self.indent) + '\n'
        patt = indexedpar.pattern.accept(self)
        seq = indexedpar.sequence.accept(self)
        return intro + proc + kiltera_syn.forkwd.string + ' ' + patt + ' ' \
            + kiltera_syn.inkwd.string + ' ' + seq + '\n'
    
    def visit_loop(self, loop, *args, **kwargs):
        intro = kiltera_syn.sequence.string + '\n'
        proc = indent(loop.process.accept(self), self.indent) + '\n'
        patt = loop.pattern.accept(self)
        seq = loop.sequence.accept(self)
        return intro + proc + kiltera_syn.forkwd.string + ' ' + patt + ' ' \
            + kiltera_syn.inkwd.string + ' ' + seq + '\n'

    def visit_interrupt(self, interrupt, *args, **kwargs):
        intro = kiltera_syn.do.string
        base = interrupt.base_process.accept(self)
        alt  = interrupt.alt_process.accept(self)
        return intro + '\n' \
            + indent(base, self.indent) + '\n' \
            + kiltera_syn.interrupt.string + '\n' \
            + indent(alt, self.indent) + '\n'

    def visit_wait(self, wait, *args, **kwargs):
        intro = kiltera_syn.wait.string
        delay_expr = wait.delay.accept(self)
        return intro + ' ' + delay_expr + ' ' + kiltera_syn.arrow.string + ' \n' \
            + indent(wait.continuation.accept(self), self.indent) + '\n'
    
    def visit_timeout(self, timeout, *args, **kwargs):
        intro = kiltera_syn.timeout.string 
        base = timeout.base_process.accept(self)
        expr = timeout.time_limit.accept(self)
        alt  = timeout.alt_process.accept(self)
        return intro + '\n' \
            + indent(base, self.indent) + '\n' \
            + kiltera_syn.after.string + ' ' + expr + ' ' + kiltera_syn.arrow.string + '\n' \
            + indent(alt, self.indent) + '\n'
    
    def visit_sendat(self, sendat, *args, **kwargs):
        return sendat.action.accept(self) + ' ' \
            + kiltera_syn.at.string + ' ' \
            + str(sendat.elapsed) + ' ' \
            + kiltera_syn.arrow.string + ' \n' \
            + indent(sendat.continuation.accept(self), self.indent) + '\n'
        
    def visit_receiveat(self, receiveat, *args, **kwargs):
        return receiveat.action.accept(self) + ' ' \
            + kiltera_syn.at.string + ' ' \
            + str(receiveat.elapsed) + ' ' \
            + kiltera_syn.arrow.string + ' \n' \
            + indent(receiveat.continuation.accept(self), self.indent) + '\n'


    # Toplevel modules
    
    def visit_funcdef(self, funcdef, *args, **kwargs):
        intro = kiltera_syn.function.string
        separator = kiltera_syn.comma.string + ' '
        param_list = separator.join(funcdef.parameters)
        param_list = kiltera_syn.lpar.string + param_list + kiltera_syn.rpar.string
        funcdef_text = intro + ' ' + str(funcdef.name) + ' ' \
            + param_list + kiltera_syn.colon.string + '\n' \
            + indent(funcdef.expression.accept(self), self.indent) + '\n'
        return funcdef_text
    
    def visit_processdef(self, processdef, *args, **kwargs):
        intro = kiltera_syn.process.string
        separator = kiltera_syn.comma.string + ' '
        port_list = separator.join(processdef.ports)
        param_list = separator.join(processdef.parameters)
        if param_list != '':
            param_list = kiltera_syn.lpar.string + param_list + kiltera_syn.rpar.string
        processdef_text = intro + ' ' + str(processdef.name) + ' ' \
            + kiltera_syn.lsqbra.string + port_list + kiltera_syn.rsqbra.string \
            + param_list + kiltera_syn.colon.string + '\n' \
            + indent(processdef.body.accept(self), self.indent) + '\n'
        return processdef_text
        

    def visit_module(self, module, *args, **kwargs):
        intro = kiltera_syn.module.string
        separator = '\n'
        imps = [kiltera_syn.importkwd.string + ' ' + name for name in module.imports]
        defs = [process_def.accept(self) for process_def in module.process_defs]
        module_text = intro + ' ' + str(module.name) 
        if len(imps) > 0:
            module_text += '\n\n'
        module_text += separator.join(imps)
        if len(defs) > 0:
            module_text += '\n\n'
        module_text += separator.join(defs)
        module_text += '\n'
        return module_text

    # Expressions
    
    def visit_bottom(self, bottom, *args, **kwargs):
        return kiltera_syn.unit.string
    
    def visit_variable(self, var, *args, **kwargs):
        return str(var.name)
    
    def visit_boolconst(self, const, *args, **kwargs):
        if bool(const.value):
            return kiltera_syn.true.string
        return kiltera_syn.false.string
    
    def visit_compare(self, comp, *args, **kwargs):
        lhs = comp.lhs.accept(self)
        rhs = comp.rhs.accept(self)
        rel = str(comp.relation)
        return lhs + ' ' + rel + ' ' + rhs
    
    def visit_boolbinexpr(self, expr, *args, **kwargs):
        left = expr.left.accept(self)
        right = expr.right.accept(self)
        op = str(expr.operator)
        if isinstance(expr.left, BoolExpr) \
        and precedence[expr.left.operator] > precedence[expr.operator]:
            left = kiltera_syn.lpar.string + left + kiltera_syn.rpar.string
        if isinstance(expr.right, BoolExpr) \
        and precedence[expr.right.operator] > precedence[expr.operator]:
            right = kiltera_syn.lpar.string + right + kiltera_syn.rpar.string
        return left + ' ' + op + ' ' + right
    
    def visit_boolunexpr(self, expr, *args, **kwargs):
        arg = expr.argument.accept(self)
        op = str(expr.operator)
        if isinstance(expr.argument, BoolExpr) \
        and precedence[expr.argument.operator] > precedence[expr.operator]:
            arg = kiltera_syn.lpar.string + arg + kiltera_syn.rpar.string
        return op + arg
        
    def visit_arithconst(self, const, *args, **kwargs):
        return str(const.value)

    def visit_arithbinexpr(self, expr, *args, **kwargs):
        left = expr.left.accept(self)
        right = expr.right.accept(self)
        op = str(expr.operator)
        if isinstance(expr.left, ArithExpr) \
        and precedence[expr.left.operator] > precedence[expr.operator]:
            left = kiltera_syn.lpar.string + left + kiltera_syn.rpar.string
        if isinstance(expr.right, ArithExpr) \
        and precedence[expr.right.operator] > precedence[expr.operator]:
            right = kiltera_syn.lpar.string + right + kiltera_syn.rpar.string
        return left + ' ' + op + ' ' + right
    
    def visit_arithunexpr(self, expr, *args, **kwargs):
        arg = expr.argument.accept(self)
        op = str(expr.operator)
        if isinstance(expr.argument, ArithExpr) \
        and precedence[expr.argument.operator] > precedence[expr.operator]:
            arg = kiltera_syn.lpar.string + arg + kiltera_syn.rpar.string
        return op + arg 
    
    def visit_string(self, string, *args, **kwargs):
        return '"' + str(string.msg) + '"'

    def visit_tuple(self, tuple_expr, *args, **kwargs):
        separator = kiltera_syn.comma.string + ' '
        items = [str(item.accept(self)) for item in tuple_expr.sequence]
        text = separator.join(items)
        if len(items) == 1:
            text += separator
        return kiltera_syn.lpar.string + text + kiltera_syn.rpar.string

    def visit_cons(self, cons_expr, *args, **kwargs):
        separator = kiltera_syn.semicolon.string + ' '
        items = [str(item.accept(self)) for item in cons_expr.sequence]
        text = separator.join(items)
        return kiltera_syn.lpar.string + text + kiltera_syn.rpar.string

    def visit_comprehension(self, comprehension, *args, **kwargs):
        text = comprehension.expression.accept(self)
        text += ' ' + kiltera_syn.forkwd.string + ' '
        text += comprehension.pattern.accept(self)
        text += ' ' + kiltera_syn.inkwd.string + ' '
        text += comprehension.sequence.accept(self)
        text += ' ' + kiltera_syn.ifkwd.string + ' '
        text += comprehension.condition.accept(self)
        return kiltera_syn.lpar.string + text + kiltera_syn.rpar.string
    
    def visit_tupleaccess(self, tupleaccess_expr, *args, **kwargs):
        tuple_expr = tupleaccess_expr.tuple_expr.accept(self)
        index_expr = tupleaccess_expr.index_expr.accept(self)
        if not isinstance(tupleaccess_expr.tuple_expr, Var) \
        and not isinstance(tupleaccess_expr.tuple_expr, TupleExpr):
            tuple_expr = kiltera_syn.lpar.string + tuple_expr + kiltera_syn.rpar.string
        return tuple_expr + kiltera_syn.lsqbra.string + index_expr + kiltera_syn.rsqbra.string
    
    def visit_funccall(self, funccall, *args, **kwargs):
        separator = ', '
        arg_vals = [arg.accept(self) for arg in funccall.args]
        actual_args = kiltera_syn.lpar.string + separator.join(arg_vals) + kiltera_syn.rpar.string
        return funccall.var.accept(self) + actual_args 
    
    def visit_newchan(self, newchan, *args, **kwargs):
        if newchan.channel_class is Async:
            chcls = kiltera_syn.async.string
        elif newchan.channel_class is Sync:
            chcls = kiltera_syn.sync.string
        return chcls + ' ' + kiltera_syn.channelkwd.string
        
    def visit_case(self, case_expr, *args, **kwargs):
        intro = kiltera_syn.cases.string
        separator = '\n' + kiltera_syn.bar.string + ' '
        then = ' ' + kiltera_syn.arrow.string + ' \n'
        expr_text = case_expr.expression.accept(self)
        cases = []
        for pattern, process in case_expr.cases:
            cases += [pattern.accept(self) + \
                        then + \
                        indent(process.accept(self), 2 + self.indent)]
        cases_text = separator.join(cases)
        return intro + ' ' + expr_text + ' ' + kiltera_syn.lbrace.string + '\n' + ' ' * self.indent + cases_text + '\n' + kiltera_syn.rbrace.string 

    def visit_condexpr(self, cond, *args, **kwargs):
        intro = kiltera_syn.ifkwd.string
        cond_text = cond.condition.accept(self)
        act_text = indent(cond.action.accept(self), self.indent)
        alt_text = indent(cond.alternative.accept(self), self.indent)
        return intro + ' ' + cond_text + ' ' + kiltera_syn.then.string + ' \n' \
            + act_text + '\n' + kiltera_syn.elsekwd.string + '\n' + alt_text + '\n'

    def visit_letexpr(self, let, *args, **kwargs):
        intro = kiltera_syn.let.string
        separator = '\n' + kiltera_syn.andop.string + ' '
        binding_list = [str(name) + ' = ' + str(let.bindings[name].accept(self)) \
                        for name in let.bindings]
        bindings = separator.join(binding_list)
        return intro + ' ' + bindings + '\n' + kiltera_syn.inkwd.string + '\n' \
            + indent(let.process.accept(self), self.indent) + '\n'

 