#! /usr/bin/python

"""
script: klt
Author: Ernesto Posse
Created on: July 15, 2006
Last update: Jan 22, 2007

Description:
  This is the kiltera interpreter.

Typical use:

  On the command line type
  
    klt [options] <file>
    
  To see the options, run
  
    klt -h

"""

import sys

try:
    if production:
        import kiltera
        sys.path.append(kiltera.__path__[0])
        version = kiltera.version
        from kiltera import *
    else:
        import os.path
        file_dir, this_file = os.path.split(__file__)
        parent_dir, this_dir = os.path.split(file_dir)
        sys.path.append(os.path.abspath('..'))
        sys.path.append(os.path.abspath(file_dir))
        sys.path.append(os.path.abspath(parent_dir))
        import kiltera
        version = kiltera.version
except NameError:
    import os.path
    file_dir, this_file = os.path.split(__file__)
    parent_dir, this_dir = os.path.split(file_dir)
    sys.path.append(os.path.abspath('..'))
    sys.path.append(os.path.abspath(file_dir))
    sys.path.append(os.path.abspath(parent_dir))
    import kiltera
    version = kiltera.version

from posse_tools.utils import *
from kiltera_pretty_printer import *
from kiltera_ast_printer import *
from interpreter import *
from simulator import *
from tracing import *

import getopt
from aperiot.parsergen import *
from aperiot.llparser import ParsingException

original_syntax      = True
print_tokens         = False
print_parsing_table  = False
print_parse_tree     = False
print_ast            = False
pretty_print         = False
use_prebuilt_parser  = True
trace_file_name      = None
trace_format         = 'std'
trace_time_precision = '.15'
main_process         = 'main'
simulate             = False
real_time_simulation = False
remote_server        = None
port                 = 50007         # Arbitrary non-privileged port; must be the same as the server.
verbose              = False
verbose_level        = ''
run = not print_parsing_table and not print_tokens and not print_parse_tree and not print_ast and not pretty_print 

def execute(file_name):
    try:
        set_verbose_level()
        source = open_file(file_name)
        parser = get_kiltera_parser()
        if print_parsing_table:
            show_parsing_table(parser)
        if print_tokens:
            source = show_tokens(source, parser)
        tree = parse(source, parser)
        source.close()
        if print_parse_tree:
            show_parse_tree(tree)
        ast = build_ast(parser, tree)
        if print_ast:
            show_ast(ast)
        if pretty_print:
            do_pretty_print(ast)
        if run:
            do_run(ast)
    except ParsingException, e:
        print_time('* (Parsing error) ')
        source.close()
        text = file(file_name, 'r')
        lines = text.readlines()
        text.close()
        try:
            e.pprint(lines[e.linenum-1])
        except:
            print e
    except KltException, e:
        print_time('* (Execution error) '+str(e))
        source.close()
    except IOError, e:
        print_time("* (I/O error) " + str(e))
    

def set_verbose_level():
    log.enabled = verbose
    level_mask = verbose_level.split(':')
    if len(level_mask) == 1:
        level, mask = (level_mask[0], 'all')
    else:
        level, mask = level_mask
    log.set_verbose_level(level, mask=mask)

def open_file(file_name):
    log.note(0, "* Openning file '" + file_name + "'")
    source = file(file_name, 'r')
    log.note(0, "* File '" + file_name + "' opened")
    return source

def get_kiltera_parser():
    log.note(0, "* Loading kiltera parser")
    syntax_module = 'kiltera_syn'
    if original_syntax:
        syntax_module = 'kiltera_syn_original'
        log.note(0, "* Using original syntax")
    if use_prebuilt_parser:
        parser = load_parser(syntax_module, verbose=verbose)
    else:
        parser = build_parser(syntax_module, verbose=verbose)
    log.note(0, "* kiltera parser loaded")
    return parser

def show_tokens(source, parser):
    log.note(0, "* Printing token stream")
    token_stream = parser.lexer.tokenize(source)
    for token in token_stream:
        print token
    source.seek(0)
    log.note(0, "* Token stream printed")
    return source

def parse(source, parser):
    log.note(0, "* Parsing begins")
    tree = parser.parse(source, apply_actions=False)
    log.note(0, "* Parsing succeded")
    return tree

def show_parsing_table(parser):
    log.note(0, "* Printing parsing table")
    print parser.cfg
    print
    print parser.table
    log.note(0, "* Parsing table printed")

def show_parse_tree(tree):
    log.note(0, "* Printing parse tree")
    print str(tree)
    log.note(0, "* Parse tree printed")

def build_ast(parser, tree):
    log.note(0, "* Building AST")
    ast = parser.apply_actions(tree)
    log.note(0, "* AST built")
    return ast
  
def show_ast(ast):
    log.note(0, "* Printing AST")
    astp = ASTPrinter()
    astp.pprint(ast)
    log.note(0, "* Finished printing AST")

def do_pretty_print(ast):
    log.note(0, "* Pretty printing")
    pp = PrettyPrinter()
    pp.pprint(ast)
    log.note(0, "* Finished pretty-printing")

def do_run(ast):
    trace_handler = None
    if trace_file_name is not None:
        log.note(0, "* Opening file for tracing: '" + str(trace_file_name) + "'")
        trace_handler = TraceHandler(trace_file_name, trace_format, trace_time_precision)    
    log.note(0, "* Creating interpreter")
    if simulate:
        interpreter = Simulator(real_time_simulation, log)
    else:
        interpreter = Interpreter(server=remote_server, port=port)
    if trace_handler:
        interpreter.register(trace_handler)
    log.note(0, "* Beginning execution")
    interpreter.execute(ast, main=main_process)
    log.note(0, "* Finished execution")
    if trace_handler is not None:
        log.note(0, "* Closing trace file: '%s'" % trace_file_name)
        trace_handler.finalize()

def usage():
    print "Usage: klt [options] <file>"
    print "where <file> must be a .klt file,"
    print "and [options] is a list of one of the following options:"
    print
    print "  General options:"
    print "  -h, --help     Prints this message."
    print "  -V             Print version number."
    print "  -v<level>      Verbose mode; shows various internal messages with the given\n" \
        + "                 level of detail. <level> is of the form all, <number>, \n" \
        + "                 <number>:<mask> or all:<mask>, where <mask> is a combination \n" \
        + "                 of: o (only one level), T (time), t (thread), m (module), \n" \
        + "                 c (class), f (function/method) and l (line number)."
    print
    print "  Execution options:"
    print "  -m<name>       Use <name> as the main process of the module."
    print "  -r<server>     Connect to a remote kiltera server (give symbolic server name.)"
    print "  -r<srv>:<prt>  Connect to a remote kiltera server <srv> using port <prt>. \n" \
       +  "                 The port number must be the same as that of the server."
    #~ print "  -s             Simulate (run in logical time rather than real-time.)"
    #~ print "  -S             Execute without threads (event-based simulation with physical \n" \
        #~ + "                 clock.)"
    print
    print "  Trace-releted options:"
    print "  -e             Print event trace."
    print "  -E<file>       Save event trace to file."
    print "  -x             Print event trace in short XML format."
    print "  -X<file>       Save event trace to file in short XML format."
    print "  -y             Print event trace in long XML format."
    print "  -Y<file>       Save event trace to file in long XML format."
    print "  -P<len>.<dec>  Format of time in the trace: total length <len> and number\n" \
        + "                 of decimals <dec>."
    print
    print "  Syntax-related options:"
    print "  -a             Print abstract syntax tree."
    print "  -b             Build kiltera parser at initialization instead of \n" \
        + "                 loading pre-built parser."
    print "  -l             Print token stream produced by lexical analysis."
    print "  -o             Use alternative syntax (nesting with indentation instead of braces.)"
    print "  -p             Pretty-print module."
    print "  -t             Print parse tree."
    print "  -T             Print parsing table."

def print_version():
    print version

def main():
    try:
        options, args = getopt.getopt(sys.argv[1:], 'abeE:lm:opP:r:RsStTv:VxX:yY:h',['help'])
        global verbose, verbose_level, use_prebuilt_parser, run, original_syntax
        global print_tokens, print_parsing_table, print_parse_tree, print_ast, pretty_print
        global trace_file_name, trace_format, trace_time_precision, main_process, simulate, real_time_simulation
        global remote_server, port
        for option, arg in options:
            if option == '-a':
                print_ast = True
            if option == '-b': 
                use_prebuilt_parser = False
            if option in ('-e','-E'):
                trace_file_name = 'stdout'
                if option == '-E':
                    trace_file_name = arg
            if option in ('-x','-X'):
                trace_file_name = 'stdout'
                trace_format = 'xmls'
                if option == '-X':
                    trace_file_name = arg
            if option in ('-y','-Y'):
                trace_file_name = 'stdout'
                trace_format = 'xmll'
                if option == '-Y':
                    trace_file_name = arg
            if option == '-l':
                print_tokens = True
            if option == '-m':
                main_process = arg
            if option == '-o':
                original_syntax = False
            if option == '-p':
                pretty_print = True
            if option == '-P':
                trace_time_precision = arg
            if option == '-r':
                parts = arg.split(':')
                remote_server = parts[0]
                if len(parts) > 1:
                    port = int(parts[1])
            if option == '-R':
                import socket
                remote_server = socket.gethostname()
            if option in ('-s', '-S'):
                simulate = True
                if option == '-S':
                    real_time_simulation = True
            if option == '-t':
                print_parse_tree = True
            if option == '-T':
                print_parsing_table = True
            if option == '-v':
                verbose = True
                verbose_level = arg
            if option == '-V':
                print_version()
                sys.exit(2)
            if option in ('-h', '--help'):
                usage()
                sys.exit(2)
        if len(args) < 1 and option not in ['-h','--help','-V','-T']:
            raise getopt.GetoptError("No file given")
        file_name = args[0]
        run = not print_parsing_table \
              and not print_tokens      \
              and not print_parse_tree  \
              and not print_ast         \
              and not pretty_print 
        execute(file_name)
    except getopt.GetoptError, e:
        print_time("* (Command-line error) " + str(e))
        usage()


class KltException(Exception):
    
    def __init__(self, msg):
        self.msg = msg
    
    def __str__(self):
        return str(self.msg)


if __name__ == "__main__":
    main()
