"""
Project: kiltera 
Module: kltserver
Author: Ernesto Posse
Created on: Nov 27, 2006
Last modified: Jan 22, 2007

Description:
  This implements a server to handle a distributed kiltera system.
"""

import sys

try:
    if production:
        import kiltera
        sys.path.append(kiltera.__path__[0])
        version = kiltera.version
        from kiltera import *
    else:
        import os.path
        dir, this_file = os.path.split(__file__)
        parent_dir, this_dir = os.path.split(dir)
        sys.path.append(os.path.abspath('..'))
        sys.path.append(os.path.abspath(dir))
        sys.path.append(os.path.abspath(parent_dir))
        import kiltera
        version = kiltera.version
except NameError:
    import os.path
    dir, this_file = os.path.split(__file__)
    parent_dir, this_dir = os.path.split(dir)
    sys.path.append(os.path.abspath('..'))
    sys.path.append(os.path.abspath(dir))
    sys.path.append(os.path.abspath(parent_dir))
    import kiltera
    version = kiltera.version

import getopt
import socket
import pickle
from threading import *
from time import *
from values import *
import random
import posse_tools
from posse_tools.utils import *


log  = Verbose()
HOST = socket.gethostname()
PORT = 50007              # Arbitrary non-privileged port

def set_verbose_level(verbose_level):
    log.enabled = verbose_level is not None
    if log.enabled:
        level_mask = verbose_level.split(':')
        if len(level_mask) == 1:
            level, mask = (level_mask[0], 'all')
        else:
            level, mask = level_mask
        log.set_verbose_level(level, mask=mask)


class Server:
    
    def __init__(self, host=HOST, port=PORT):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((host, port))
        self.clients_by_module = {}
        self.clients_by_host = {}
    
    def accept_clients(self):
        log.note(1, "1 -- begin accepting clients")
        while True:
            log.note(1, "2 -- waiting for a client")
            conn, addr = self.socket.accept()
            host, port = addr
            host_name = socket.gethostbyaddr(host)[0]
            log.note(1, "3 -- new client -- %s" % host_name)
            request_listener = Thread(target=self.client_handler, args=(conn,addr,host_name))
            request_listener.start()
        log.note(1, "4 -- finnished")
    
    def client_handler(self, socket, addr, host_name):
        log.note(2, "1 -- new client handler -- %s -- %s -- %s" % (host_name, socket, repr(addr)))
        module_name = ''
        try:
            module_name = self.register_client(socket, addr, host_name)
            log.note(2, "2 -- new client registered -- %s -- %s -- %s -- %s" % (module_name, host_name, socket, repr(addr)))
            while True:
                data = socket.recv(2**16)
                log.note(2, "3 -- message from client -- %s -- %s -- %s" % (module_name,host_name,data[:10]))
                if not data:
                    break
                self.process_message(data, module_name)
        except ServerException, e:
            print e
        except Exception, e:
            print e
        log.note(2, "4 -- client stops -- %s -- %s -- %s" % (module_name,host_name,data[:10]))
        socket.close()
        self.remove_connection(module_name, host_name, socket, addr)
        log.note(2, "5 -- client connection removed -- %s -- %s -- %s" % (module_name,host_name,data[:10]))

    def register_client(self, socket, addr, host_name):
        log.note(3, "1 -- registering client -- %s -- %s -- %s" % (host_name, socket, repr(addr)))
        module_name = socket.recv(1024)
        if not self.valid_name(module_name):
            socket.send('invalid name')
            raise ServerException('Client with invalid name: "%s"' % module_name)
        self.add_connection(module_name, host_name, socket, addr)
        log.note(3, "2 -- added connection -- %s -- %s -- %s -- %s" % (host_name, module_name, socket, repr(addr)))
        socket.send('acknowledged')
        log.note(3, "3 -- sent acknowledgement -- %s -- %s -- %s -- %s" % (host_name, module_name, socket, repr(addr)))
        return module_name
    
    def valid_name(self, s):
        if len(s) == 0:
            return False
        if not s[0].isalpha():
            return False
        return all(c.isalnum() or c == '_' for c in s)
            
    def process_message(self, request, module_name=None):
        log.note(4, "1 -- processing message -- %s -- %s" % (module_name, request[:10]))
        try:
            unpacked_request = pickle.loads(request)
            log.note(4, "2 -- unpacked message -- %s -- %s" % (module_name, repr(unpacked_request)))
            target_name, msg = unpacked_request
            log.note(4, "3 -- extracted target name -- %s -- %s" % (module_name, repr(target_name)))
            target_socket, target_addr = self.get_connection(target_name)
            log.note(4, "4 -- obtained target connection -- %s -- %s" % (module_name, repr(target_name)))
            repacked = pickle.dumps(msg) 
            log.note(4, "5 -- repacked message -- %s -- %s" % (module_name, repr(target_name)))
            target_socket.send(repacked)
            log.note(4, "6 -- sent message to target -- %s -- %s" % (module_name, repr(target_name)))
        except TypeError:
            raise ServerException('Client sent a message with incorrect format: %s -- %s' % (module_name, request[:10]))
        except KeyError:
            raise ServerException('Client sent a message with incorrect target: %s -- %s' % (module_name, request[:10]))
        log.note(4, "7 -- done processing message -- %s -- %s" % (module_name, request))

    def add_connection(self, module_name, host_name, socket, addr):
        if module_name in self.clients_by_module:
            if host_name in self.clients_by_module[module_name]:
                self.clients_by_module[module_name][host_name].append((socket, addr))
            else:
                self.clients_by_module[module_name][host_name] = [(socket, addr)]
        else:
            self.clients_by_module[module_name] = {host_name: [(socket, addr)]}
        if host_name in self.clients_by_host:
            if module_name in self.clients_by_host[host_name]:
                self.clients_by_host[host_name][module_name].append((socket, addr))
            else:
                self.clients_by_host[host_name][module_name] = [(socket, addr)]
        else:
            self.clients_by_host[host_name] = {module_name: [(socket, addr)]}
                
    def get_connection(self, target_name):
        parts = target_name.split(':')
        if len(parts) == 0:
            raise ServerException("incorrect target (empty)")
        if len(parts) == 1:
            if parts[0] == '':
                raise ServerException("incorrect target (empty)")
            module_name = parts[0]
            hosts = self.clients_by_module[module_name].keys()
            host_name = random.choice(hosts)
        else:
            module_name = parts[1]
            host_name = parts[0]
        try:
            socket, addr = random.choice(self.clients_by_module[module_name][host_name])
            return (socket, addr)
        except KeyError, e:
            raise ServerException("connection for '%s' not found" % target_name)
        except IndexError, e:
            raise ServerException("connection for '%s' not found" % target_name)
    
    def remove_connection(self, module_name, host_name, socket, addr):
        self.clients_by_module[module_name][host_name].remove((socket, addr))

    def run(self):
        #~ client_listener  = Thread(target=self.accept_clients)
        #~ client_listener.start()
        self.socket.listen(5)
        self.accept_clients()


class ServerException(Exception):
    
    def __init__(self, msg):
        self.msg = msg
    
    def __str__(self):
        return "kiltera server exception: %s" % self.msg


def usage():
    print "usage: kltserver [options]"
    print "where [options] is a list of one of the following options:"
    print "  -p<int>        Use the given port for client connections."
    print "  -v<level>      Verbose mode; shows various internal messages with the given\n" \
        + "                 level of detail. <level> is of the form all, <number>, \n" \
        + "                 <number>:<mask> or all:<mask>, where <mask> is a combination \n" \
        + "                 of: o (only one level), T (time), t (thread), m (module), \n" \
        + "                 c (class), f (function/method) and l (line number)."
    print "  -V             Print version number."
    print "  -h, --help     Prints this message."

def print_version():
    print version


def main():
    try:
        options, args = getopt.getopt(sys.argv[1:], 'p:v:Vh',['help'])
        global PORT
        level = None
        for option, arg in options:
            if option == '-p':
                PORT = int(arg)
            if option == '-v':
                level = arg
            if option == '-V':
                print_version()
                sys.exit(2)
            if option in ('-h', '--help'):
                usage()
                sys.exit(2)
        set_verbose_level(level)
        log.note(1,'* Server begins -- host: %s' % HOST)
        server = Server()
        server.run()
    except getopt.GetoptError, e:
        print_time("* (Command-line error) " + str(e))
        usage()
    except ValueError, e:
        print_time("* (Command-line error) Invalid port number.")
    except KeyboardInterrupt:
        print_time("* Server interrupted")

if __name__ == '__main__':
    main()
