"""
Project: kiltera 
Module: kthreads
Author: Ernesto Posse
Created on: Aug 15, 2006
Last modified: Jan 22, 2007

Description:
  This implements a class of "killable" threads and some supporting resources such as events and
  binary semaphores.
  
  The general scheme can be summarized as follows: We implement a new class of threads called
  KThread. KThreads are hierarchical, that is a thread can have "subthreads." A thread checks 
  explicitly at key points whether some other thread has ordered it to die. We call these
  life check-points "traps."
  
  If a thread is ordered to die, the thread kills all its subthreads and performs any cleanup 
  operations by executing a callback function.
  
  A thread may be blocked at certain points, usually when trying to acquire a resource. In
  order to kill a blocked thread it is necessary to wake it up so that it can perform the life
  check. To do this, before blocking, the thread stores a reference to the resource where it 
  may block. When other thread kills the first, it notifies the blocking resource to
  wake up its blocked threads so that the killed thread performs the life-check and dies.
  
  A thread may in fact have more than one blocking resource associated to it. This is useful
  in certain situations where other threads may be blocked by a resource held by the thread
  being killed.
  
  To support this scheme, we reimplement the Event class from the standard threading module and we 
  implement a binary semaphore, similar to the BoundedSemaphore from the standard threading module.
  These reimplementation are made such that the Event.wait and BinarySemaphore.acquire methods,
  where a thread may block, register themselves with the thread so that they may be woken up if
  needed, and perform the life-check right after the blocking point. 
  
  We also reimplement the join method of the Thread class to support this "trap" protocol, and therefore
  we replace the standard run method by one which, when the thread finishes, sets its __alive flag to
  false. We also provide a "recursive_join" which joins the thread and all its subthreads recursively.
  
  To implement join, the calling thread may block, but it does not need to register the join's target
  thread as a blockin resource because the finish method executed when the target ends will wake up
  the blocked threads.
  
  Some details:
 
  * Threads have an internal flag __alive which is set to false when the thread is killed.
  * Threads also have a list __blocking_resources which is meant to contain the object where
    the thread is blocked (if it is blocked,) as well as any other "observer" objects which 
    may need to be awaken when the thread is unblocked. The objects in this list must have a
    "wakeup" method (with no arguments.)
  * Whenever a thread enters a critical section of a resource where it might block, it registers 
    the resource in its __blocking_resources list, and when the thread finishes the critical section
    it removes the resource from the list.
  * When a thread is killed, it sets its __alive flag to false, it recursively kills any subthreads
    and it wakes up all sleeping resources so that they "fall into the trap" (see next point.)
  * In a critical section where a thread may block, the first line immediately after the point where
    if may block (e.g. a wait,) the thread executes its check_life method, passing as parameter an
    optional callback function used for cleanup operations. This check_life method checks the thread's   
    __alive flag. If __alive is false, the thread calls the optional callback function and ends by 
    raising a SystemExit exception. (This mechanism is what I call "falling into the trap.")

  Finally we provide a "hardkill" method which first invokes the normal kill method but then resets the
  localtrace function for the thread (and recursively for the subthreads) so that the thread dies in the 
  next line executed. This approach does not require the killed threads to call check_life, and because 
  of this, no cleanup actions are taken.
"""

import sys
#~ import ctypes
from threading import *
from kexceptions import *
from posse_tools.utils import log

def _async_raise(tid, excobj):
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(excobj))
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        # """if it returns a number greater than one, you're in trouble, 
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemError("PyThreadState_SetAsyncExc failed")
 

class KBinarySemaphore:
    
    counter = 0
    
    def __init__(self, name=None, green=True):
        KBinarySemaphore.counter += 1
        if name is None:
            self.name = 'sem' + str(KBinarySemaphore.counter)
        else:
            self.name = name
        self.green = green
        self.owner = None
        self.lock = RLock(verbose=False)
        self.condition = Condition(self.lock,verbose=False)
    
    def __repr__(self):
        return "%s(%s,%s,%s)" % (self.__class__.__name__, self.name, self.green, self.owner)
        
    def acquire(self, blocking=True, external_lock=None, reset_sems=[], release_sems=[]):
        success = False
        self.condition.acquire()
        me = currentThread()
        me.register_blocking_resource(self)
        cleanup_data = {'condition_lock':self.condition,
                        'external_lock' :external_lock, 
                        'reset_sems'    :reset_sems, 
                        'release_sems'  :release_sems}
        log.note(5, "1 -- %s -- %s"%(me,self))
        while not self.green:
            log.note(5, "2 -- %s -- %s"%(me,self))
            if not blocking:
                break
            self.condition.wait()
            log.note(5, "3 -- %s -- %s"%(me,self))
            currentThread().check_life(cleanup_func=self.death_cleanup, 
                                       cleanup_kwargs=cleanup_data)
            log.note(5, "4 -- %s -- %s"%(me,self))
        else:
            self.green = False
            self.owner = me
            success = True
            log.note(5, "5 -- %s -- %s"%(currentThread(),self))
        me.remove_blocking_resource(self)
        self.condition.release()
        log.note(5, "6 %s -- %s"%(currentThread(),self))
        return success
        
    def release(self):
        log.note(5, "1 -- %s -- %s"%(currentThread(),self))
        self.condition.acquire()
        log.note(5, "2 -- %s -- %s"%(currentThread(),self))
        self.green = True
        self.owner = None
        self.condition.notify()
        log.note(5, "3 -- %s -- %s"%(currentThread(),self))
        self.condition.release()
        log.note(5, "4 -- %s -- %s"%(currentThread(),self))
    
    def reset(self, green=True):
        self.condition.acquire()
        log.note(5, "1 -- %s -- %s"%(currentThread(),self))
        self.green = green
        log.note(5, "2 -- %s -- %s"%(currentThread(),self))
        self.condition.release()
    
    def wakeup(self):
        log.note(5, "1 -- %s -- %s"%(currentThread(),self))
        self.condition.acquire()
        log.note(5, "2 -- %s -- %s"%(currentThread(),self))
        self.condition.notifyAll()
        log.note(5, "3 -- %s -- %s"%(currentThread(),self))
        self.condition.release()
        log.note(5, "4 -- %s -- %s"%(currentThread(),self))
    
    def death_cleanup(self, condition_lock=None, external_lock=None, reset_sems=[], release_sems=[]):
        me = currentThread()
        log.note(5, "1 -- %s -- %s"%(me,self))
        for external_sem, flag in reset_sems:
            external_sem.reset(flag)
            log.note(5, "2 -- %s -- %s"%(me,self))
        for external_sem in release_sems:
            external_sem.release()
            log.note(5, "3 -- %s -- %s"%(me,self))
        if external_lock is not None:
            external_lock.release()
            log.note(5, "4 -- %s -- %s"%(me,self))
        #~ self.condition_lock.notifyAll() # Needed?
        self.condition.release()
        log.note(5, "5 -- %s -- %s" % (me,self))



class KEvent:

    counter = 0

    def __init__(self):
        KEvent.counter += 1
        self.id = KEvent.counter
        self.__cond = Condition(Lock())
        self.__flag = False
        
    def __repr__(self):
        return "%s(%d,%s)" % (self.__class__.__name__, self.id, self.__flag)

    def isSet(self):
        return self.__flag

    def set(self):
        log.note(4, "1 -- %s" % self)
        self.__cond.acquire()
        log.note(4, "2 -- %s" % self)
        try:
            self.__flag = True
            self.__cond.notifyAll()
            log.note(4, "3 -- %s" % self)
        finally:
            self.__cond.release()
        log.note(4, "4 -- %s" % self)

    def clear(self):
        log.note(4, "1 -- %s" % self)
        self.__cond.acquire()
        log.note(4, "2 -- %s" % self)
        self.__flag = False
        log.note(4, "3 -- %s" % self)
        self.__cond.release()
        log.note(4, "4 -- %s" % self)

    def wait(self, timeout=None):
        me = currentThread()
        t = timeout or -1.0
        log.note(4, "1 -- %s -- %f" % (self,t))
        self.__cond.acquire()
        log.note(4, "2 -- %s -- %f" % (self,t))
        me.register_blocking_resource(self)
        # Note that the following is an if and not a while, so waking up the thread amounts to the same as setting the flag to true or the timeout passing.
        if not self.__flag:
            log.note(4, "3 -- %s -- %f" % (self,t))
            self.__cond.wait(timeout)
            log.note(4, "4 -- %s -- %f" % (self,t))
            currentThread().check_life(cleanup_func=self.death_cleanup)
        log.note(4, "5 -- %s -- %f" % (self, t))
        me.remove_blocking_resource(self)
        log.note(4, "6 -- %s -- %f" % (self,t))
        self.__cond.release()
        log.note(4, "7 -- %s -- %f" % (self,t))
    
    def wakeup(self):
        log.note(4, "1 -- %s" % self)
        self.__cond.acquire()
        log.note(4, "2 -- %s" % self)
        self.__cond.notifyAll()
        log.note(4, "3 -- %s" % self)
        self.__cond.release()
        log.note(4, "4 -- %s" % self)
    
    def death_cleanup(self):
        self.__cond.notifyAll()
        self.__cond.release()


class KThread(Thread):

    def __init__(self, *args, **keywords):
        Thread.__init__(self, *args, **keywords)
        self.__alive = True
        self.target = keywords['target']
        self.__subthreads = []
        self.__blocking_resources = []
        self.__life_condition = Condition()
        self.__stderr = sys.stderr
    
    def __repr__(self):
        return "%s(%s,%s,%s,%d,%s)" % (self.__class__.__name__, self.getName(), self.__alive, self.isAlive(), len(self.subthreads), self.target)
    
    def __str__(self):
        return repr(self)
    
    def tree_str(self):
        s = repr(self) + '\n'
        if len(self.__subthreads) > 0:
            s += '  |\n'
        lines = []
        i = 0
        while i < len(self.__subthreads):
            thread = self.__subthreads[i]
            t = thread.tree_str()
            lines = t.splitlines()
            if i < len(self.__subthreads) - 1:
                lines = ['  +- ' + lines[0]] \
                      + ['  |  ' + line for line in lines[1:]]
            else:
                lines = ['  +- ' + lines[0]] \
                      + ['     ' + line for line in lines[1:]]            
            i += 1
            s += '\n'.join(lines) + '\n'
        if len(self.__subthreads) == 0:
            return s + '\n'
        return s
        

    def get_subthreads(self):
        return self.__subthreads
    
    def set_subthreads(self, subthreads):
        self.__subthreads = subthreads
    
    def del_subthreads(self):
        for thread in self.__subthreads:
            thread.kill()
    
    subthreads = property(get_subthreads, set_subthreads, del_subthreads)
    
    def get_some_children_alive(self):
        return any(thread.__alive for thread in self.__subthreads)
    
    some_children_alive = property(get_some_children_alive)

    def register_blocking_resource(self, resource):
        self.__blocking_resources.append(resource)
        log.note(3, "1 -- %s -- %s" % (self,resource))

    def remove_blocking_resource(self, resource):
        try:
            self.__blocking_resources.remove(resource)
            log.note(3, "1 -- %s -- %s" % (self,resource))
        except ValueError:
            # If resource is not in the list, just ignore
            log.note(3, "2 -- %s -- %s" % (self,resource))

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        Thread.start(self)

    def __run(self):
        try:
            self.__run_backup()
            #~ for thread in self.__subthreads:
                #~ thread.recursive_join()
        except RuntimeException, e:
            print>>self.__stderr, str(e)
        except StopThread, e:
            print>>self.__stderr, str(e)
        self.run = self.__run_backup
        self.finish()   # Frees up any blocked resources waiting on the thread.

    def __stop(self):
        #~ log.note(3, "1 -- %s" % self)
        self.__life_condition.acquire()
        #~ log.note(3, "2 -- %s" % self)
        self.__alive = False
        self.__life_condition.notifyAll()
        #~ log.note(3, "3 -- %s" % self)
        self.__life_condition.release()
        #~ log.note(3, "4 -- %s" % self)

    def finish(self):
        log.note(3, "1 -- %s" % self)
        self.__stop()
        log.note(3, "2 -- %s" % self)
        for resource in self.__blocking_resources:
            resource.wakeup()
        log.note(3, "3 -- %s" % self)
        if hasattr(self, 'keyboard_handler'):
            self.keyboard_handler.alive = False

    def kill(self):
        log.note(3, "1 -- %s" % self)
        self.__stop()
        for thread in self.__subthreads:
            thread.kill()
        log.note(3, "2 -- %s" % self)
        # We need to wake up blocked resources so that they fall into the killing trap when they execute the next check_life call.
        for resource in self.__blocking_resources:
            resource.wakeup()
        log.note(3, "3 -- %s" % self)
    
    def __die(self, cleanup_func=None, cleanup_args=[], cleanup_kwargs={}):
        log.note(3, "1 -- %s"%self)
        if self != currentThread():
            log.note(3, "2 -- %s is being killed by %s"%(self,currentThread()))
        if callable(cleanup_func):
            log.note(3, "3 -- %s"%self)
            cleanup_func(*cleanup_args, **cleanup_kwargs)
        log.note(3, "thread dead -- %s"%self)
        raise SystemExit()
    
    def check_life(self, cleanup_func=None, cleanup_args=[], cleanup_kwargs={}):
        log.note(3, "1 -- %s"%self)
        self.__life_condition.acquire()
        log.note(3, "2 -- %s"%self)
        if not self.__alive:
            self.__life_condition.release()
            log.note(3, "3 -- %s"%self)
            self.__die(cleanup_func, cleanup_args, cleanup_kwargs)
        log.note(3, "4 -- %s"%self)
        self.__life_condition.release()

    def hardkill(self):
        log.note(3, "1 -- %s" % self)
        self.kill()
        log.note(3, "2 -- %s" % self)
        self.__settrace()
        #~ self.terminate()
        log.note(3, "3 -- %s" % self)

    def __recursive_settrace(self):
        log.note(2, "1 -- %s" % self)
        for thread in self.__subthreads:
            log.note(2, "2 -- %s" % self)
            thread.__recursive_settrace()
        log.note(2, "3 -- %s" % self)
        sys.settrace(self.globaltrace) 
        log.note(2, "4 -- %s" % self)
        
    def __settrace(self):
        log.note(2, "1 -- %s" % self)
        sys.settrace(self.globaltrace) 

    def globaltrace(self, frame, why, arg):
        log.note(2, "1 -- %s" % self)
        if why == 'call':
            log.note(2, "2 -- %s" % self)
            return self.localtrace
        else:
            log.note(2, "3 -- %s" % self)
            return None

    def localtrace(self, frame, why, arg):
        log.note(2, "1 -- %s" % self)
        if not self.__alive:
            log.note(2, "2 -- %s" % self)
            if why == 'line':
                log.note(3, "1 -- %s" % self)
                raise SystemExit()
        return self.localtrace

    def join(self):
        me = currentThread()
        # 'self' is the target thread of the join, i.e. the thread that 'me' (aka the current thread) is waiting to finish.
        assert me is not self, "cannot join current thread"
        def death_cleanup(lock):
            #~ lock.notifyAll() # Needed?
            lock.release()
        log.note(3, "1 -- %s -- %s" % (self,me))
        #~ log.traceback(3)
        self.__life_condition.acquire()
        log.note(3, "2 -- %s -- %s" % (self,me))
        #~ self.register_blocking_resource(me)  # Not needed if the target ends with a call to finish
        while self.__alive:
            log.note(3, "3 -- %s -- %s" % (self,me))
            self.__life_condition.wait()
            log.note(3, "4 -- %s -- %s" % (self,me))
            me.check_life(cleanup_func=death_cleanup, 
                          cleanup_args=[self.__life_condition])
        log.note(3, "5 -- %s -- %s" % (self,me))
        #~ self.remove_blocking_resource(me)  # Not needed if the target ends with a call to finish
        self.__life_condition.release()
        log.note(3, "6 -- %s -- %s" % (self,me))

    def recursive_join(self):
        for subthread in self.__subthreads:
            subthread.recursive_join()
        self.join()
        
    def wakeup(self):
        log.note(3, "1 -- %s -- %s"%(currentThread(),self))
        self.__life_condition.acquire()
        log.note(3, "2 -- %s -- %s"%(currentThread(),self))
        self.__life_condition.notifyAll()
        log.note(3, "3 -- %s -- %s"%(currentThread(),self))
        self.__life_condition.release()
        log.note(3, "4 -- %s -- %s"%(currentThread(),self))

    def raise_exc(self, excobj):
        assert self.isAlive(), "thread must be started"
        for tid, tobj in threading._active.items():
            if tobj is self:
                _async_raise(tid, excobj)
                return
        
        # the thread was alive when we entered the loop, but was not found 
        # in the dict, hence it must have been already terminated. should we raise
        # an exception here? silently ignore?
    
    def terminate(self):
        # must raise the SystemExit type, instead of a SystemExit() instance
        # due to a bug in PyThreadState_SetAsyncExc
        self.raise_exc(SystemExit)
