"""
Project: kiltera 
Module: kthreads
Author: Ernesto Posse
Created on: Oct 13, 2006
Last modified: Jan 22, 2007

Description:
  This implements a class of "killable" process instances for the kiltera simulator.
"""

from posse_tools.utils import all
from posse_tools.stacks import *

class ProcessInstance(object):
    
    counter = 0
    
    def __init__(self, process, parent=None):
        ProcessInstance.counter += 1
        self.id = ProcessInstance.counter 
        self.process = process
        self.__subprocesses = []
        self.__finished = False 
        self.__interacted = False
        self.parent = parent   # Parent should be another ProcessInstance
        self.continuation_stack = Stack()   # Stack of Continuation objects
        self.events = []
    
    def __repr__(self):
        return "%s(%s, %s, %s, %d, %s)" % (self.__class__.__name__, self.id, self.__alive, len(self.__subprocesses), self.process)
    
    def __str__(self):
        return repr(self)
        
    def tree_str(self):
        s = repr(self) + '\n'
        if len(self.__subprocesses) > 0:
            s += '  |\n'
        lines = []
        i = 0
        while i < len(self.__subprocesses):
            subprocess = self.__subprocesses[i]
            t = subprocess.tree_str()
            lines = t.splitlines()
            if i < len(self.__subprocesses) - 1:
                lines = ['  +- ' + lines[0]] \
                      + ['  |  ' + line for line in lines[1:]]
            else:
                lines = ['  +- ' + lines[0]] \
                      + ['     ' + line for line in lines[1:]]            
            i += 1
            s += '\n'.join(lines) + '\n'
        if len(self.__subprocesses) == 0:
            return s + '\n'
        return s

    def get_subprocesses(self):
        return self.__subprocesses
    
    def set_subprocesses(self, subprocesses):
        self.__subprocesses = subprocesses
        for subprocess in subprocesses:
            subprocess.parent = self
    
    def add_subprocess(self, subprocess):
        self.__subprocesses.append(subprocess)
        subprocess.parent = self
    
    #~ def del_subprocesses(self):
        #~ for subprocess in self.__subprocesses:
            #~ subprocess.kill()
    
    subprocesses = property(get_subprocesses, set_subprocesses) #, del_subprocesses)
    
    def get_finished(self):
        if self.__finished:
            return True
        self.__finished = all(component.finished for component in self.__subprocesses)  #Maybe this could be optimized by simply keeping a counter of live children.
        return self.__finished
    
    def set_finished(self, val):
        self.__finished = val
        if self.__finished and not self.continuation_stack.isempty():
            continuation = self.continuation_stack.pop()
            continuation()
        if self.parent is not None:
            self.parent.__update_finished()
        
    def __update_finished(self):
        """This method is meant to be invoked by the child processes when they finish only."""
        self.__finished = all(component.finished for component in self.__subprocesses)
        if self.__finished and not self.continuation_stack.isempty():
            continuation = self.continuation_stack.pop()
            continuation()
        if self.__finished and self.parent is not None:
            self.parent.__update_finished()
    
    finished = property(get_finished, set_finished)
    
    def get_interacted(self):
        if self.__interacted:
            return True
        self.__interacted = any(component.interacted for component in self.__subprocesses)  #Maybe this could be optimized by simply keeping a counter of live children.
        return self.__interacted
    
    def set_interacted(self, val):
        self.__interacted = val
        if self.parent is not None:
            self.parent.__update_interacted()
        
    def __update_interacted(self):
        """This method is meant to be invoked by the child processes when they interact only."""
        self.__interacted = any(component.interacted for component in self.__subprocesses)
        if self.parent is not None:
            self.parent.__update_interacted()
    
    interacted = property(get_interacted, set_interacted)
    
    def add_continuation(self, cont):
        self.continuation_stack.push(cont)
    
    def add_event(self, event):
        self.events.append(event)
    
        
class Continuation:
    
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
    
    def __call__(self):
        self.func(*self.args, **self.kwargs)
