"""
Project: kiltera 
Module: simulator
Author: Ernesto Posse
Created on: Sep 18, 2006
Last modified: Jan 22, 2007

Description:
  This implements an event-based simulator of kiltera processes.
"""

from kiltera_ast import *
from values import *
from environments import *
from sim_channels import *
from ktime import *
from event_sched import *
from pinstances import *
from visitor import *
from posse_tools.utils import *
from builtins import *
import copy

class Simulator(Visitor):
    
    def __init__(self, physical_clock=False, verbose=False):
        self.observers = []
        if physical_clock:
            self.clock = PhysicalClock()
        else:
            self.clock = LogicalClock()
        self.event_scheduler = EventScheduler(self.clock, verbose)
        self.event_scheduler.visitor = self
    
    def execute(self, module, main=None):
        new_env = Env(Frame(self.builtins()))
        module.accept(self, new_env, main=main)
        self.event_scheduler.run()
    
    def register(self, observer):
        self.observers.append(observer)
    
    def update(self, *args):
        for observer in self.observers:
            observer.notify(*args)
    
    # Atom visitors
    
    def visit_atom(self, atom, *args, **kwargs):
        env, inst = args[0], args[1]
        
    def visit_output(self, output, *args, **kwargs):
        env, inst = args[0], args[1]
        channel = output.channel_ref.accept(self, env)
        data = output.data.accept(self, env)
        self.update(time=self.clock.time, kind='ext', action='send', channel=output.channel_ref, data=data, loc=output.parent.get_full_location(), pos=output.parent.pos)
        return (channel, data)
        
    def visit_input(self, input_action, *args, **kwargs):
        env, inst = args[0], args[1]
        channel = input_action.channel_ref.accept(self, env)
        self.update(time=self.clock.time, kind='ext', action='receive', channel=input_action.channel_ref, data=data, loc=input_action.parent.get_full_location(), pos=input_action.parent.pos)
        return channel
        
    def visit_assign(self, assignment, *args, **kwargs):
        env, inst = args[0], args[1]
        try:
            actual_value = assignment.expression.accept(self, env)
            env[assignment.variable] = actual_value
            self.update(time=self.clock.time, kind='int/ext', action='assign', var=assignment.variable, data=actual_value, loc=assignment.parent.get_full_location(), pos=assignment.parent.pos)
        except EnvironmentException:
            raise RuntimeException(assignment.parent, "assignment error") 
        
    def visit_print(self, print_atom, *args, **kwargs):
        env, inst = args[0], args[1]
        msg = print_atom.msg.accept(self, env)
        print str(msg)
        self.update(time=self.clock.time, kind='int', action='print', data=msg, loc=print_atom.parent.get_full_location(), pos=print_atom.parent.pos)
        
    
    
    # Process visitors
    
    def visit_process(self, process, *args, **kwargs):
        env, inst = args[0], args[1]
        
    def visit_nil(self, nil, *args, **kwargs):
        env, inst = args[0], args[1]
        log.note(2, "1 -- %s", nil)

    def visit_done(self, nil, *args, **kwargs):
        env, inst = args[0], args[1]
        inst.finished = True
        log.note(2, "1 -- %s", done)

    def visit_prefix(self, prefix, *args, **kwargs):
        log.note(2, "1 -- %s", prefix)
        env, inst = args[0], args[1]
        self.event_scheduler.enter_last(0, prefix.action, self, *args, **kwargs)
        self.event_scheduler.enter_last(0, prefix.continuation, self, *args, **kwargs)
        #~ prefix.continuation.accept(self, *args, **kwargs)
        
    def visit_send(self, send, *args, **kwargs):
        log.note(2, "1 -- %s", send)
        env, inst = args[0], args[1]
        channel, data = send.action.accept(self, *args, **kwargs)
        time = self.clock.time
        channel.send(data)
        if len(channel.receivers) > 0:
            if channel.policy == RANDOM:
                i = randint(0, len(channel.receivers) - 1)
            else:
                i = 0
            chosen, recv_args, recv_kwargs = channel.receivers[i]
            del channel.receivers[i]
            self.event_scheduler.enter_last(0, chosen, *recv_args, **recv_kwargs)
        inst.interacted = True
        self.update(time, 'ext', 'send', data, send.get_full_location(), send.pos)
        self.event_scheduler.enter_last(0, send.continuation, self, *args, **kwargs)
        
    def visit_receive(self, receive, *args, **kwargs):
        log.note(2, "1 -- %s", receive)
        env, inst = args[0], args[1]
        channel = receive.action.accept(self, *args, **kwargs)
        if channel.data_available():
            data = channel.receive()
            inst.interacted = True
            self.update(self.clock.time, 'ext', 'receive', data, receive.get_full_location(), receive.pos)
            new_env = Env(Frame({receive.parameter: data}), env)
            self.event_scheduler.enter_last(0, receive.continuation, self, new_env, *args[1:], **kwargs)
        else:
            channel.receivers.append((receive, ((self,) + args), kwargs))
        
    def visit_parallel(self, parallel, *args, **kwargs):
        log.note(2, "1 -- %s", parallel)
        env, inst = args[0], args[1]
        for subprocess in parallel.process_list:
            subprocess_inst = ProcessInstance(subprocess)
            inst.add_subprocess(subprocess_inst)
            new_args = [env] + [subprocess_inst] + args[2:]
            self.event_scheduler.enter_last(0, subprocess, self, *new_args, **kwargs)
            #~ self.event_scheduler.enter_rand(0, process.accept, self, *args, **kwargs)
        
    def visit_sequence(self, sequence, *args, **kwargs):
        log.note(2, "1 -- %s", sequence)
        env, inst = args[0], args[1]
        first = sequence.process_list[0]
        remainder = Sequence(sequence.process_list[1:], sequence.pos)
        new_continuation = Continuation(remainder, self, *args, **kwargs)
        inst.add_continuation(new_continuation)
        self.event_scheduler.enter_last(0, first, self, *args, **kwargs)
        
        #~ for subprocess in sequence.process_list:
            #~ value = process.accept(self, *args)
            #~ if isinstance(value, ExecutingProcess):
                #~ self.join(value.threads)

    def visit_local(self, local, *args, **kwargs):
        log.note(2, "1 -- %s", local)
        env, inst = args[0], args[1]
        if local.channel_class is Sync:
            channel_class = SyncChannel
        else:
            channel_class = AsyncChannel
        table = dict((name, channel_class()) for name in local.channel_names)
        new_env = Env(Frame(table), env)
        return local.process.accept(self, new_env, *args[1:])

    def visit_chanarray(self, chanarray, *args, **kwargs):
        log.note(2, "1 -- %s", chanarray)
        env, inst = args[0], args[1]
        name = chanarray.name
        size = chanarray.size.accept(self, env)
        if chanarray.channel_class is Sync:
            channel_class = SyncChannel
        else:
            channel_class = AsyncChannel
        array = [channel_class() for i in range(size)]
        new_binding = dict([(name, Tuple(array))])
        new_env = Env(Frame(new_binding), env)
        return chanarray.process.accept(self, new_env, *args[1:])

    def visit_let(self, let, *args, **kwargs):
        log.note(2, "1 -- %s", let)
        env, inst = args[0], args[1]
        bindings = dict((var, let.bindings[var].accept(self, env)) 
                        for var in let.bindings)
        new_env = Env(Frame(bindings), env)
        return let.process.accept(self, new_env, *args[1:])

    def visit_conditional(self, cond, *args, **kwargs):
        log.note(2, "1 -- %s", cond)
        env, inst = args[0], args[1]
        condition_value = cond.condition.accept(self, env).value
        if condition_value is True:
            return cond.action.accept(self, *args)
        else:
            return cond.alternative.accept(self, *args)

    def visit_match(self, match_proc, *args, **kwargs):
        log.note(2, "1 -- %s", match_proc)
        env, inst = args[0], args[1]
        expr_val = match_proc.expression.accept(self, env)
        for pattern, process in match_proc.cases:
            corresponds, subst = match(pattern, expr_val)
            if corresponds:
                new_env = Env(Frame(subst), env)
                return process.accept(self, new_env, *args[1:])
        return Bottom
        
    def visit_procinst(self, process_inst, *args, **kwargs):
        log.note(2, "1 -- %s", process_inst)
        env, inst = args[0], args[1]
        latent_process = env[process_inst.process_name]
        processdef = latent_process.processdef
        actual_ports = [port.accept(self,env) 
                        for port in process_inst.actual_ports]
        port_bindings = map(lambda a, b: (a,b), 
                            processdef.ports, actual_ports)
        argument_values = [arg.accept(self, env) 
                            for arg in process_inst.actual_params]
        param_bindings = map(lambda a, b: (a,b), 
                            processdef.parameters, argument_values)
        new_bindings = dict(port_bindings + param_bindings)
        new_env = Env(Frame(new_bindings), latent_process.env)
        new_instance = copy.deepcopy(processdef.body)
        new_instance.parent = process_inst.parent
        process_inst.parent.add_subinstance(new_instance)
        return new_instance.accept(self, new_env, *args[1:])
    
    def visit_localdefs(self, localdefs, *args, **kwargs):
        log.note(2, "1 -- %s", localdefs)
        env, inst = args[0], args[1]
        new_env = Env(Frame(), env)
        for d in localdefs.defs:
            d.accept(self, new_env)
        return localdefs.body.accept(self, new_env, *args[1:])
        
    def visit_indexedpar(self, indexedpar, *args, **kwargs):
        log.note(2, "1 -- %s", indexedpar)
        env, inst = args[0], args[1]
        process = indexedpar.process
        pattern = indexedpar.pattern
        sequence = indexedpar.sequence
        threads = []
        seq_obj = sequence.accept(self, env)
        if not isinstance(seq_obj, Tuple):
            raise ExecutionException()
        for item in seq_obj.value:
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                new_args = tuple([self, new_env] + list(args[1:]))
                threads += [KThread(target = process, args = new_args)]
        indexedpar.threads = threads
        for thread in threads:
            thread.start()
        self.join(threads)
        
    def visit_loop(self, loop, *args, **kwargs):
        log.note(2, "1 -- %s", loop)
        env, inst = args[0], args[1]
        process = loop.process
        pattern = loop.pattern
        sequence = loop.sequence
        seq_obj = sequence.accept(self, env)
        if not isinstance(seq_obj, Tuple):
            raise ExecutionException()
        for item in seq_obj.value:
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                value = process.accept(self, new_env, *args[1:])

    def visit_interrupt(self, interrupt, *args, **kwargs):
        log.note(2, "1 -- %s", interrupt)
        env, inst = args[0], args[1]
        signal = Event()
        base_thread = KThread(target = interrupt.base_process, args = tuple([self] + list(args)))
        alt_thread = KThread(target = interrupt.alt_process, args = (self, env, signal))
        class BaseController(Process):
            def __call__(self):
                base_thread.start()
                base_thread.join()
                if alt_thread.isAlive():
                    alt_thread.kill()
                if alt_controller.isAlive():
                    alt_controller.kill()
        class AltController(Process):
            def __call__(self):
                alt_thread.start()
                signal.wait()
                if signal.isSet() and base_thread.isAlive():
                    base_thread.kill()
                    self.base_controller.kill()
        base_controller = BaseController()
        alt_controller = AltController()
        base_controller_thread = KThread(target = base_controller)
        alt_controller_thread = KThread(target = alt_controller)
        base_controller.alt_controller = alt_controller_thread
        alt_controller.base_controller = base_controller_thread
        interrupt.threads = [base_thread, alt_thread, base_controller_thread, alt_controller_thread]
        base_controller_thread.start()
        alt_controller_thread.start()
        return None
    
    def visit_wait(self, wait, *args, **kwargs):
        log.note(2, "1 -- %s", wait)
        env, inst = args[0], args[1]
        delay_value = wait.delay.accept(self, env)
        self.event_scheduler.enter_last(float(delay_value), wait.continuation, self, *args, **kwargs)
    
    def visit_timeout(self, timeout, *args, **kwargs):
        log.note(2, "1 -- %s", timeout)
        env, inst = args[0], args[1]
        self.event_scheduler.enter_last(0, timeout.base_process, self, *args, **kwargs)
        def decider():
            #~ print timeout.base_process.finished
            #~ print timeout.base_process.interacted
            if not timeout.base_process.finished and not timeout.base_process.interacted:
                #~ print '**********************************cancelling'
                # remove all events associated with the base process from the event list
                timeout.base_process.cancel(self.event_scheduler)
                #~ print 'base cancelled'
                # execute the alternative
                self.event_scheduler.enter_last(0, timeout.alt_process, self, *args, **kwargs)
                #~ print 'alt scheduled'
        time_limit = float(timeout.time_limit.accept(self, env))
        self.event_scheduler.enter_last(time_limit, decider)
    
    def visit_sendat(self, sendat, *args, **kwargs):
        log.note(2, "1 -- %s", sendat)
        env, inst = args[0], args[1]
        #~ time_before_action = self.clock.time
        #~ sendat.action.accept(self, env)
        #~ time_after_action = self.clock.time
        #~ elapsed_time = Number(time_after_action - time_before_action)
        #~ if len(args) > 1:
            #~ signal = args[1]
            #~ signal.set()
        #~ new_frame = Frame({sendat.elapsed: elapsed_time})
        #~ new_env = Env(new_frame, env)
        #~ sendat.continuation.accept(self, new_env, *args[1:])

        channel, data = sendat.action.accept(self, *args, **kwargs)
        #~ self.event_scheduler.enter_last(0, send.action.accept, self, *args, **kwargs)
        time_before_action = self.clock.time
        channel.send(data)
        time_after_action = self.clock.time
        if len(channel.receivers) > 0:
            if channel.policy == RANDOM:
                i = randint(0, len(channel.receivers) - 1)
            else:
                i = 0
            chosen, recv_args, recv_kwargs = channel.receivers[i]
            del channel.receivers[i]
            self.event_scheduler.enter_last(0, chosen, *recv_args, **recv_kwargs)
        elapsed_time = Number(time_after_action - time_before_action)
        sendat.interacted = True
        self.update(time, 'ext', 'send', data, sendat.get_full_location(), sendat.pos)
        #~ if len(args) > 1:
            #~ signal = args[1]
            #~ signal.set()
        new_frame = Frame({sendat.elapsed: elapsed_time})
        new_env = Env(new_frame, env)
        self.event_scheduler.enter_last(0, sendat.continuation, self, new_env, *args[1:], **kwargs)

    def visit_receiveat(self, receiveat, *args, **kwargs):
        log.note(2, "1 -- %s", receiveat)
        env, inst = args[0], args[1]
        now = self.clock.time
        channel = receiveat.action.accept(self, *args, **kwargs)
        if channel.data_available():
            data = channel.receive()
            if hasattr(receiveat, 'entry_time'):
                receive_state_entry_time = receiveat.entry_time
                del receiveat.entry_time
            else:
                receive_state_entry_time = now
            #~ if len(args) > 1:
                #~ signal = args[1]
                #~ signal.set()
            elapsed_time = Number(now - receive_state_entry_time)
            new_frame = Frame({receiveat.parameter: data, receiveat.elapsed: elapsed_time})
            new_env = Env(new_frame, env)
            receiveat.interacted = True
            self.update(self.clock.time, 'ext', 'receive', data, receiveat.get_full_location(), receiveat.pos)
            self.event_scheduler.enter_last(0, receiveat.continuation, self, new_env, *args[1:], **kwargs)
        else:
            receiveat.entry_time = now
            channel.receivers.append((receiveat, ((self,) + args), kwargs))
    
    # Toplevel modules

    def visit_funcdef(self, funcdef, *args, **kwargs):
        env, inst = args[0], args[1]
        env.add_binding(funcdef.name, LatentFunction(funcdef, env))

    def visit_processdef(self, processdef, *args, **kwargs):
        env, inst = args[0], args[1]
        env.add_binding(processdef.name, LatentProcess(processdef, env))
    
    def visit_module(self, module, *args, **kwargs):
        env, inst = args[0], args[1]
        new_env = Env(Frame(), env)
        for processdef in module.process_defs:
            processdef.accept(self, new_env)
        if 'main' in kwargs and kwargs['main'] is not None:
            main = kwargs['main']
        else:
            main = 'main'
        main_proc = new_env[main]
        main_inst = ProcessInstantiation(main, [], [])
        main_inst.accept(self, new_env, **kwargs)

    # Expression
    
    def visit_bottom(self, bottom, *args, **kwargs):
        return Bottom
    
    def visit_variable(self, var, *args, **kwargs):
        env, inst = args[0], args[1]
        result = env[var.name]
        return result
    
    def visit_boolconst(self, const, *args, **kwargs):
        return Boolean(const.value)

    def visit_compare(self, comp, *args, **kwargs):
        env, inst = args[0], args[1]
        lhs = comp.lhs.accept(self, env).value
        rhs = comp.rhs.accept(self, env).value
        rel = comp.relation
        return Boolean(rel(lhs, rhs))
    
    def visit_boolbinexpr(self, expr, *args, **kwargs):
        env, inst = args[0], args[1]
        left = expr.left.accept(self, env).value
        right = expr.right.accept(self, env).value
        op = expr.operator
        return Boolean(op(left, right))

    def visit_boolunexpr(self, expr, *args, **kwargs):
        env, inst = args[0], args[1]
        arg = expr.argument.accept(self, env).value
        op = expr.operator
        return Boolean(op(arg))
        
    def visit_arithconst(self, const, *args, **kwargs):
        return Number(const.value)
        
    def visit_arithbinexpr(self, expr, *args, **kwargs):
        env, inst = args[0], args[1]
        left = expr.left.accept(self, env).value
        right = expr.right.accept(self, env).value
        op = expr.operator
        return Number(op(left, right))

    def visit_arithunexpr(self, expr, *args, **kwargs):
        env, inst = args[0], args[1]
        arg = expr.argument.accept(self, env).value
        op = expr.operator
        return Number(op(arg))
        
    def visit_string(self, string, *args, **kwargs):
        return String(string.msg)

    def visit_tuple(self, tuple_expr, *args, **kwargs):
        env, inst = args[0], args[1]
        seq = tuple(item.accept(self, env) for item in tuple_expr.sequence)
        return Tuple(seq)
    
    def visit_comprehension(self, comprehension, *args, **kwargs):
        env, inst = args[0], args[1]
        expression = comprehension.expression
        pattern = comprehension.pattern
        sequence = comprehension.sequence
        condition = comprehension.condition
        outcome = []
        for item in sequence.accept(self, env).value:
            #~ success, subst = match(pattern, item.accept(self, env)) 
            success, subst = match(pattern, item) 
            if success:
                new_env = Env(Frame(subst), env)
                if condition.accept(self, new_env).value:
                    outcome += [expression.accept(self, new_env)]
        return Tuple(outcome)
    
    def visit_tupleaccess(self, tupleaccess_expr, *args, **kwargs):
        env, inst = args[0], args[1]
        tuple_val = tupleaccess_expr.tuple_expr.accept(self, env).value
        index = tupleaccess_expr.index_expr.accept(self, env).value
        return tuple_val[index]
    
    def visit_funccall(self, funccall, *args, **kwargs):
        env, inst = args[0], args[1]
        latent_func = funccall.var.accept(self, env)
        argument_values = [arg.accept(self, env) 
                            for arg in funccall.args]
        if isinstance(latent_func, BuiltInFunction):
            pyargs = [arg.pyvalue() for arg in argument_values]
            result = latent_func(*pyargs)
            klt_res = PyValue(result).klt_value
            return klt_res
        else:
            funcdef = latent_func.funcdef
            param_bindings = map(lambda a, b: (a,b), 
                                 funcdef.parameters, argument_values)
            new_bindings = dict(param_bindings)
            new_env = Env(Frame(new_bindings), latent_func.env)
            return funcdef.expression.accept(self, new_env)
        
    
    def visit_newchan(self, newchan, *args, **kwargs):
        if newchan.channel_class is Async:
            chcls = 'au'
        elif newchan.channel_class is Sync:
            chcls = 'su'
        return Channel(chcls)

    


    # Other methods

    def join(self, threads):
        if threads is not None:
            for thread in threads:
                thread.join()
    
    def builtins(self):
        return setup_built_ins()

############################################################################
            


def match(pattern, datum, subst=None):
    if subst is None:
        subst = {}
    if isinstance(pattern, ConstantExpression):
        return pattern.value == datum.value, subst
    elif isinstance(pattern, Variable):
        if pattern.name in subst and subst[pattern.name] != datum:
            print "Mismatch: '" + pattern.name + "' was previously matched with '" + str(subst[pattern.name]) + "'"
            return False, {}
        subst[pattern.name] = datum
        return True, subst
    elif isinstance(pattern, TupleExpression):
        if not isinstance(datum, Tuple) \
        or len(pattern.sequence) != len(datum.value):
            return False, subst
        correspond, new_subst = True, subst
        i = 0
        while i < len(pattern.sequence) and correspond:
            correspond, new_subst = match(pattern.sequence[i], datum.value[i], new_subst)
            i += 1
        return correspond, new_subst
    return False, subst

class ExecutionException(Exception):
    pass


