"""
Project: kiltera 
Module: tracing
Author: Ernesto Posse
Created on: Sep 8, 2006
Last modified: Jan 22, 2007

Description:
  This implements a tracer for kiltera processes.
"""
  
import sys

from posse_tools.utils import indent
from kiltera_pretty_printer import *

dtdl = \
    '<!DOCTYPE trace [\n' \
    + '  <!ELEMENT trace (event*)>\n' \
    + '  <!ELEMENT event (time, kind, action, channel?, data?, location, pos?)>\n' \
    + '  <!ELEMENT time (#PCDATA)>\n' \
    + '  <!ELEMENT kind (#PCDATA)>\n' \
    + '  <!ELEMENT action (#PCDATA)>\n' \
    + '  <!ELEMENT channel (#PCDATA)>\n' \
    + '  <!ELEMENT var (#PCDATA)>\n' \
    + '  <!ELEMENT data (#PCDATA)>\n' \
    + '  <!ELEMENT location (#PCDATA)>\n' \
    + '  <!ELEMENT pos (#PCDATA)>\n' \
    + ']>\n'

dtds = \
    '<!DOCTYPE trace [\n' \
    + '  <!ELEMENT trace (event*)>\n' \
    + '  <!ELEMENT event EMPTY>\n' \
    + '  <!ATTLIST event time CDATA #REQUIRED>\n' \
    + '  <!ATTLIST event kind CDATA #REQUIRED>\n' \
    + '  <!ATTLIST event action CDATA #REQUIRED>\n' \
    + '  <!ATTLIST event channel CDATA>\n' \
    + '  <!ATTLIST event var CDATA>\n' \
    + '  <!ATTLIST event data CDATA>\n' \
    + '  <!ATTLIST event location CDATA #REQUIRED>\n' \
    + '  <!ATTLIST event pos CDATA>\n' \
    + ']>\n'

xml_header = \
    '<?xml version="1.0" encoding="UTF-8"?>\n'

class TraceHandler:
  
    def __init__(self, trace_file_name = 'stdout', kind='std', time_format=10):
        self.event_list = []
        self.pp = PrettyPrinter()
        self.time_format = '%' + time_format + 'f'
        global xml_header, dtds, dtdl
        if trace_file_name == 'stdout':
            self.trace_file = sys.stdout
            xml_header = ''
            dtds = ''
            dtdl = ''
        else:
            self.trace_file = file(trace_file_name, 'w')
        self.kind = kind
        self.header = \
            {
                'std': 'start-of-trace',
                'xmls': xml_header + dtds + '<trace>',
                'xmll': xml_header + dtdl + '<trace>'
            }
        self.footer = \
            {
                'std': 'end-of-trace',
                'xmls': '</trace>',
                'xmll': '</trace>'
            }
        self.evt_formatter = \
            {
                'std' : self.pformat_event,
                'xmls' : self.xml_short_element,
                'xmll' : self.xml_long_element
            }
        self.initialize()
    
    def initialize(self):
        line = self.header[self.kind] + '\n'
        self.trace_file.write(line)
    
    def finalize(self):
        line = self.footer[self.kind] + '\n'
        self.trace_file.write(line)
        self.trace_file.close()
    
    def flush(self):
        self.trace_file.flush()
    
    def notify(self, *args, **kwargs):
        self.event_list.append(kwargs)
        evt_text = self.evt_formatter[self.kind](*args, **kwargs)
        evt_text = indent(evt_text, 2) + '\n'
        self.trace_file.write(evt_text)
    
    def pformat_event(self, *args, **kwargs):
        time       = kwargs['time']
        evt_kind   = kwargs['kind']
        evt_action = kwargs['action']
        if evt_action in ['send','receive']:
            channel = kwargs['channel']
        elif evt_action == 'assign':
            var = kwargs['var']
        data = kwargs['data']
        loc = kwargs['loc']
        pos = kwargs['pos']
        fields = []
        fields += ['time = %.'+self.time_precision+'s' % time]
        fields += ['kind = %s' % evt_kind]
        fields += ['action = %s' % evt_action]
        if evt_action in ['send','receive']:
            fields += ['channel = %s' % self.pp.pformat(channel)]
        elif evt_action == 'assign':
            fields += ['var = %s' % var.name]
        fields += ['data = %s' % str(data)]
        if len(loc) == 0:
            loc_text = 'main'
        else:
            loc_text = '.'.join([str(proc.name) for proc in loc])
        fields += ['location = %s' % loc_text]
        fields += ['pos = (%d,%d)' % pos[:2]]
        text = ', '.join(fields)
        return '(' + text + ')'

    def xml_long_element(self, *args, **kwargs):
        time       = kwargs['time']
        evt_kind   = kwargs['kind']
        evt_action = kwargs['action']
        if evt_action in ['send','receive']:
            channel = kwargs['channel']
        elif evt_action == 'assign':
            var = kwargs['var']
        data = kwargs['data']
        loc = kwargs['loc']
        pos = kwargs['pos']
        fields = []
        fields += ['<time>%s</time>' % time]
        fields += ['<kind>"%s"</kind>' % evt_kind]
        fields += ['<action>"%s"</action>' % evt_action]
        if evt_action in ['send','receive']:
            fields += ['<channel>"%s"</channel>' % self.pp.pformat(channel)]
        elif evt_action == 'assign':
            fields += ['<var>"%s"</var>' % var.name]
        fields += ['<data>"%s"</data>' % str(data)]
        if len(loc) == 0:
            loc_text = 'main'
        else:
            loc_text = '.'.join([str(proc.name) for proc in loc])
        fields += ['<location>"%s"</location>' % loc_text]
        fields += ['<pos>"(%d,%d)"</pos>' % pos[:2]]
        text = '\n'.join([' ' * indent + field for field in fields])
        text = '<event>' + '\n' + text + '\n' + '</event>'
        return text

    def xml_short_element(self, *args, **kwargs):
        time       = kwargs['time']
        evt_kind   = kwargs['kind']
        evt_action = kwargs['action']
        if evt_action in ['send','receive']:
            channel = kwargs['channel']
        elif evt_action == 'assign':
            var = kwargs['var']
        data = kwargs['data']
        loc = kwargs['loc']
        pos = kwargs['pos']
        fields = []
        fields += [('time = '+self.time_format) % time]
        fields += ['kind = "%s"' % evt_kind]
        fields += ['action = "%s"' % evt_action]
        if evt_action in ['send','receive']:
            fields += ['channel = "%s"' % self.pp.pformat(channel)]
        elif evt_action == 'assign':
            fields += ['var = "%s"' % var.name]
        fields += ['data = "%s"' % str(data)]
        if len(loc) == 0:
            loc_text = 'main'
        else:
            loc_text = '.'.join([str(proc.name) for proc in loc])
        fields += ['location = "%s"' % loc_text]
        fields += ['pos = "(%d,%d)"' % pos[:2]]
        text = ' '.join(fields)
        text = '<event ' + text + '>'
        return text
