"""
Project: kiltera 
Module: values
Author: Ernesto Posse
Created on: Jun 12, 2006
Last modified: Aug 29, 2006

Description:
  This module provides the definition of the values of kiltera processes, i.e. the possible values that variables can be assigned. This includes constants, latent processes (processes that can be instantiated), and channels (defined in the channels module.)
"""

from types import *

class Value:
    def pyvalue(self):
        return self.value

class BottomVal(Value):
    def __init__(self):
        self.value = None
    def __str__(self):
        return 'None'


class Boolean(Value):
    def __init__(self, value):
        self.value = bool(value)
    def __str__(self):
        return str(self.value)


class Number(Value):
    def __init__(self, value):
        self.value = value
    def __float__(self):
        return float(self.value)
    def __int__(self):
        return int(self.value)
    def __str__(self):
        return str(self.value)


class String(Value):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return str(self.value)


class Tuple(Value):
    def __init__(self, value):
        self.value = value
    def pyvalue(self):
        return [item.pyvalue() for item in self.value]
    def __str__(self):
        l = [str(item) for item in self.value]
        if len(l) == 1:
            return '(' + l[0] + ',' + ')'
        return '(' + ', '.join(l) + ')'
        

class LatentProcess(Value):
    
    def __init__(self, processdef, env):
        self.processdef = processdef
        self.env = env
    
    def __str__(self):
        intro = 'process '
        separator = ', '
        port_list = separator.join(self.processdef.ports)
        param_list = separator.join(self.processdef.parameters)
        processdef_text = intro + str(self.processdef.name) \
            + '[' + port_list + '](' + param_list + ') = ...' 
        return processdef_text

class LatentFunction(Value):
    
    def __init__(self, funcdef, env):
        self.funcdef = funcdef
        self.env = env
    
    def __str__(self):
        intro = 'function '
        separator = ', '
        params = separator.join(self.funcdef.parameters)
        text = intro + str(self.funcdef.name) \
            + '(' + params + ') = ...'
        return text


#~ class ExecutingProcess(Value):
    
    #~ def __init__(self, threads):
        #~ self.threads = threads


class BuiltInFunction(Value):
    
    def __init__(self, pyfunc):
        self.pyfunc = pyfunc
        self.value = pyfunc
    def __call__(self, *args):
        return self.pyfunc(*args)
    def pyvalue(self):
        return self.pyfunc
    def __str__(self):
        return str(self.pyfunc)

class PyValue(Value):
    
    def __init__(self, value):
        self.value = value
        self.klt_value = self.get_klt_value(value)
    def pyvalue(self):
        return self.value
    def get_klt_value(self, value):
        import channels
        if type(value) is NoneType:
            return BottomVal()
        elif type(value) is int or type(value) is float:
            return Number(value)
        elif type(value) is str:
            return String(value)
        elif type(value) is bool:
            return Boolean(value)
        elif type(value) is FunctionType:
            return BuiltInFunction(value)
        elif type(value) is tuple or type(value) is list:
            return Tuple([self.get_klt_value(item) for item in value])
        elif type(value) is InstanceType and isinstance(value, channels.BaseChannel):
            return value
    def __str__(self):
        return str(self.value)

Bottom = BottomVal()

